"""Implement batch-level attack steps."""

import torch
import higher
import random

from ..utils import _gradient_matching, bypass_last_layer


def construct_attack(novel_defense, model, loss_fn, dm, ds, tau, init, optim, num_classes, setup):
    """Interface for this submodule."""
    eps = novel_defense['strength']  # The defense parameter encodes the eps bound used during training
    if 'adversarial-evasion' in novel_defense['type']:
        return AdversarialAttack(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-wb' in novel_defense['type']:
        return AlignmentPoisoning(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-se' in novel_defense['type']:
        return MatchingPoisoning(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-mp' in novel_defense['type']:
        return MetaPoisoning(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-fc' in novel_defense['type'] or 'adversarial-cp' in novel_defense['type']:
        return FeatureCollisionPoisoning(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-random' in novel_defense['type']:
        return RandomAttack(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-laplacian' in novel_defense['type']:
        return RandomAttack(model, loss_fn, dm, ds, tau, eps, 'laplacian', optim, num_classes, setup)
    elif 'adversarial-bernoulli' in novel_defense['type']:
        return RandomAttack(model, loss_fn, dm, ds, tau, eps, 'bernoulli', optim, num_classes, setup)
    elif 'adversarial-watermark' in novel_defense['type']:
        return WatermarkPoisoning(model, loss_fn, dm, ds, setup=setup)
    elif 'adversarial-patch' in novel_defense['type']:
        return PatchAttack(model, loss_fn, dm, ds, tau, eps, 'zero', 'none', num_classes, setup)
    elif 'adversarial-paired-patch' in novel_defense['type']:
        return PatchAttackPairs(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-variant-patch' in novel_defense['type']:
        return PatchAttackVariant(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-eps-patch' in novel_defense['type']:
        return PatchAttackVariantKnownSize(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-image-patch' in novel_defense['type']:
        return PatchAttackImageBased(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-matched-patch' in novel_defense['type']:
        return PatchAttackFixedLocation(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-adaptive-patch' in novel_defense['type']:
        return AdaptivePatchAttack(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-adaptiveV2-patch' in novel_defense['type']:
        return AdaptivePatchAttack(model, loss_fn, dm, ds, tau, eps, 'bernoulli', 'Adam', num_classes, setup)
    elif 'adversarial-adaptiveUC-patch' in novel_defense['type']:
        return AdaptivePatchAttackUnconstrained(model, loss_fn, dm, ds, tau, eps, init, optim, num_classes, setup)
    elif 'adversarial-htbd' in novel_defense['type']:
        return HTBD(model, loss_fn, dm, ds, setup=setup)
    else:
        raise ValueError(f'Invalid adversarial training objective specified: {novel_defense["type"]}.')


class BaseAttack(torch.nn.Module):
    """Implement a variety of input-altering attacks."""

    def __init__(self, model, loss_fn, dm=(0, 0, 0), ds=(1, 1, 1), tau=0.1, eps=16, init='zero',
                 optim='signAdam', num_classes=10, setup=dict(device=torch.device('cpu'), dtype=torch.float)):
        """Initialize with dict containing type and strength of attack and model info."""
        super().__init__()
        self.model = model
        self.loss_fn = loss_fn
        self.setup = setup
        self.num_classes = num_classes

        self.dm, self.ds, = dm, ds
        self.tau, self.eps = tau, eps
        self.bound = self.eps / self.ds / 255

        self.init = init
        self.optim = optim

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        if delta is None:
            delta = self._init_perturbation(inputs.shape)
        optimizer = self._init_optimizer([delta])

        for step in range(steps):
            optimizer.zero_grad()
            # Gradient step
            loss = self._objective(inputs + delta, labels, temp_targets, temp_fake_labels)
            delta.grad, = torch.autograd.grad(loss, delta, retain_graph=False, create_graph=False, only_inputs=True)
            # Optim step
            if 'sign' in self.optim:
                delta.grad.sign_()
            optimizer.step()
            # Projection step
            with torch.no_grad():
                delta.data = torch.max(torch.min(delta, self.bound), -self.bound)
                delta.data = torch.max(torch.min(delta, (1 - self.dm) / self.ds - inputs), - self.dm / self.ds - inputs)

        delta.requires_grad = False
        additional_info = None
        return delta, additional_info

    def _objective(self, inputs, labels, temp_targets, temp_fake_labels):
        raise NotImplementedError()

    def _init_perturbation(self, input_shape):
        if self.init == 'zero':
            delta = torch.zeros(input_shape, device=self.setup['device'], dtype=self.setup['dtype'])
        elif self.init == 'rand':
            delta = (torch.rand(input_shape, device=self.setup['device'], dtype=self.setup['dtype']) - 0.5) * 2
            delta *= self.eps / self.ds / 255
        elif self.init == 'bernoulli':
            delta = (torch.rand(input_shape, device=self.setup['device'], dtype=self.setup['dtype']) > 0.5).float() * 2 - 1
            delta *= self.eps / self.ds / 255
        elif self.init == 'randn':
            delta = torch.randn(input_shape, device=self.setup['device'], dtype=self.setup['dtype'])
            delta *= self.eps / self.ds / 255
        elif self.init == 'laplacian':
            loc = torch.as_tensor(0.0, device=self.setup['device'])
            scale = torch.as_tensor(self.eps / self.ds / 255, device=self.setup['device']).mean()
            generator = torch.distributions.laplace.Laplace(loc=loc, scale=scale)
            delta = generator.sample(input_shape)
        elif self.init == 'normal':
            delta = torch.randn(input_shape, device=self.setup['device'], dtype=self.setup['dtype'])
        else:
            raise ValueError(f'Invalid init {self.init} given.')

        # Clamp initialization in all cases
        delta.data = torch.max(torch.min(delta, self.eps / self.ds / 255), -self.eps / self.ds / 255)
        delta.requires_grad_()
        return delta

    def _init_optimizer(self, delta_iterable):
        tau_sgd = (self.bound * self.tau).mean()
        if 'Adam' in self.optim:
            return torch.optim.Adam(delta_iterable, lr=self.tau, weight_decay=0)
        elif 'momSGD' in self.optim:
            return torch.optim.SGD(delta_iterable, lr=tau_sgd, momentum=0.9, weight_decay=0)
        else:
            return torch.optim.SGD(delta_iterable, lr=tau_sgd, momentum=0.0, weight_decay=0)


class AdversarialAttack(BaseAttack):
    """Implement a basic untargeted attack objective."""

    def _objective(self, inputs, labels, temp_targets, temp_labels):
        """Evaluate negative CrossEntropy for a gradient ascent."""
        outputs = self.model(inputs)
        loss = -self.loss_fn(outputs, labels)
        return loss


class RandomAttack(BaseAttack):
    """Sanity check: do not actually attack - just use the random initialization."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        if delta is None:
            delta = self._init_perturbation(inputs.shape)

        # skip optimization
        pass

        delta.requires_grad = False
        return delta, None


class WatermarkPoisoning(BaseAttack):
    """Sanity check: attack by watermarking."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective. This is effectively a slight mixing.

        with mixing factor lmb = 1 - eps / 255.
        """
        img_shape = temp_targets.shape[1:]
        num_targets = temp_targets.shape[0]
        num_inputs = inputs.shape[0]

        # Place
        if num_targets == num_inputs:
            delta = temp_targets - inputs
        elif num_targets < num_inputs:
            delta = temp_targets.repeat(num_inputs // num_targets, 1, 1, 1)[:num_inputs] - inputs
        else:
            factor = num_targets // num_inputs
            delta = temp_targets[:(factor * num_targets)].reshape(num_inputs, -1, *img_shape).mean(dim=1) - inputs
        delta *= self.eps / self.ds / 255

        return delta, None


class AlignmentPoisoning(BaseAttack):
    """Implement limited steps for data poisoning via gradient alignment."""

    def _objective(self, inputs, labels, temp_targets, temp_fake_labels):
        """Evaluate Gradient Alignment and descend."""
        differentiable_params = [p for p in self.model.parameters() if p.requires_grad]

        poison_loss = self.loss_fn(self.model(inputs), labels)
        poison_grad = torch.autograd.grad(poison_loss, differentiable_params, retain_graph=True, create_graph=True)

        target_loss = self.loss_fn(self.model(temp_targets), temp_fake_labels)
        target_grad = torch.autograd.grad(target_loss, differentiable_params, retain_graph=True, create_graph=True)

        return _gradient_matching(poison_grad, target_grad)


class MatchingPoisoning(BaseAttack):
    """Implement limited steps for data poisoning via gradient alignment."""

    def _objective(self, inputs, labels, temp_targets, temp_fake_labels):
        """Evaluate Gradient Alignment and descend."""
        differentiable_params = [p for p in self.model.parameters() if p.requires_grad]

        poison_loss = self.loss_fn(self.model(inputs), labels)
        poison_grad = torch.autograd.grad(poison_loss, differentiable_params, retain_graph=True, create_graph=True)

        target_loss = self.loss_fn(self.model(temp_targets), temp_fake_labels)
        target_grad = torch.autograd.grad(target_loss, differentiable_params, retain_graph=True, create_graph=True)

        objective, tnorm = 0, 0
        for pgrad, tgrad in zip(poison_grad, target_grad):
            objective += 0.5 * (tgrad - pgrad).pow(2).sum()
            tnorm += tgrad.detach().pow(2).sum()
        return objective / tnorm.sqrt()  # tgrad is a constant normalization factor as in witch_matching


class MetaPoisoning(BaseAttack):
    """Implement limited steps for data poisoning via MetaPoison."""

    NADAPT = 2

    def _objective(self, inputs, labels, temp_targets, temp_fake_labels):
        """Evaluate Metapoison."""
        optimizer = torch.optim.SGD(filter(lambda p: p.requires_grad, self.model.parameters()), lr=0.1)
        # Wrap the model into a meta-object that allows for meta-learning steps via monkeypatching:
        with higher.innerloop_ctx(self.model, optimizer, copy_initial_weights=False) as (fmodel, fopt):
            for _ in range(self.NADAPT):
                outputs = fmodel(inputs)
                poison_loss = self.loss_fn(outputs, labels)

                fopt.step(poison_loss)

        prediction = (outputs.data.argmax(dim=1) == labels).sum()
        # model.eval()
        target_loss = self.loss_fn(fmodel(temp_targets), temp_fake_labels)
        return target_loss


class FeatureCollisionPoisoning(BaseAttack):
    """Implement limited steps for data poisoning via feature collision (with the bullseye polytope variant)."""

    def _objective(self, inputs, labels, temp_targets, temp_labels):
        """Evaluate Gradient Alignment and descend."""
        feature_model, last_layer = bypass_last_layer(self.model)

        # Get standard output:
        outputs = feature_model(inputs)
        outputs_targets = feature_model(temp_targets)

        return (outputs.mean(dim=0) - outputs_targets.mean(dim=0)).pow(2).mean()


class HTBD(BaseAttack):
    """Implement limited steps for data poisoning via hidden trigger backdoor.

    Note that this attack modifies temp_targets as a side-effect!
    """

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        if delta is None:
            delta = self._init_perturbation(inputs.shape)
        optimizer = self._init_optimizer([delta])

        temp_targets = self._apply_patch(temp_targets)
        for step in range(steps):
            input_indcs, target_indcs = self._index_mapping(inputs, temp_targets)
            optimizer.zero_grad()
            # Gradient step
            loss = self._objective(inputs + delta, temp_targets, input_indcs, target_indcs)
            delta.grad, = torch.autograd.grad(loss, delta, retain_graph=False, create_graph=False, only_inputs=True)
            # Optim step
            if 'sign' in self.optim:
                delta.grad.sign_()
            optimizer.step()
            # Projection step
            with torch.no_grad():
                delta.data = torch.max(torch.min(delta, self.bound), -self.bound)
                delta.data = torch.max(torch.min(delta, (1 - self.dm) / self.ds - inputs), - self.dm / self.ds - inputs)

        delta.requires_grad = False
        return delta, None

    def _objective(self, inputs, temp_targets, input_indcs, target_indcs):
        """Evaluate Gradient Alignment and descend."""
        feature_model, last_layer = bypass_last_layer(self.model)
        new_inputs = torch.zeros_like(inputs)
        new_targets = torch.zeros_like(temp_targets)
        for i in range(len(input_indcs)):
            new_inputs[i] = inputs[input_indcs[i]]
            new_targets[i] = temp_targets[target_indcs[i]]

        outputs = feature_model(new_inputs)
        outputs_targets = feature_model(new_targets)
        return (outputs - outputs_targets).pow(2).mean(dim=1).sum()

    def _apply_patch(self, temp_targets):
        patch_shape = [[3, random.randint(int(0.2 * temp_targets.shape[2]), int(0.4 * temp_targets.shape[2])),
                        random.randint(int(0.1 * temp_targets.shape[3]), int(0.2 * temp_targets.shape[3]))] for _ in range(temp_targets.shape[0])]

        patch = self._create_patch(patch_shape)
        patch = [p.to(**self.setup) for p in patch]
        x_locations, y_locations = self._set_locations(temp_targets.shape, patch_shape)
        for i in range(len(patch)):
            temp_targets[i, :, x_locations[i]:x_locations[i] + patch_shape[i][1], y_locations[i]:y_locations[i]
                         + patch_shape[i][2]] = patch[i] - temp_targets[i, :, x_locations[i]:x_locations[i] + patch_shape[i][1],
                                                                        y_locations[i]:y_locations[i] + patch_shape[i][2]]
        return temp_targets

    def _index_mapping(self, inputs, temp_targets):
        with torch.no_grad():
            feature_model, last_layer = bypass_last_layer(self.model)
            feat_source = feature_model(inputs)
            feat_target = feature_model(temp_targets)
            dist = torch.cdist(feat_source, feat_target)
            input_indcs = []
            target_indcs = []
            for _ in range(feat_target.size(0)):
                dist_min_index = (dist == torch.min(dist)).nonzero(as_tuple=False).squeeze()
                input_indcs.append(dist_min_index[0])
                target_indcs.append(dist_min_index[1])
                dist[dist_min_index[0], dist_min_index[1]] = 1e5
        return input_indcs, target_indcs

    def _set_locations(self, input_shape, patch_shape):
        """Fix locations where we’ll put the patches."""
        x_locations = []
        y_locations = []
        for i in range(input_shape[0]):
            x_locations.append(random.randint(0, input_shape[2] - patch_shape[i][1]))
            y_locations.append(random.randint(0, input_shape[3] - patch_shape[i][2]))
        return x_locations, y_locations

    def _create_patch(self, patch_shape):
        # create same patch or different one?
        patches = []
        for i in range(len(patch_shape)):
            temp_patch = 0.5 * torch.ones(patch_shape[i][0], patch_shape[i][1], patch_shape[i][2])
            patch = torch.bernoulli(temp_patch)
            patches.append(patch)
        return patches


class PatchAttack(BaseAttack):
    """Randomly patch 2 classes."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape = [[3, random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                        random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]

        x_locations, y_locations = self._set_locations(inputs.shape, labels, patch_shape)
        # Maybe different patch per class?
        patch = self._create_patch(patch_shape)

        if delta is None:
            delta1 = self._init_perturbation(inputs.shape)
            delta2 = self._init_perturbation(temp_targets.shape)
        delta1.requires_grad = False
        delta2.requires_grad = False

        for i in range(delta1.shape[0]):
            # Patch every class
            temp_label = labels[i]
            delta1[i, :, x_locations[i]:x_locations[i] + patch_shape[temp_label][1], y_locations[i]:y_locations[i]
                   + patch_shape[temp_label][2]] = patch[temp_label] - inputs[i, :, x_locations[i]:x_locations[i]
                                                                              + patch_shape[temp_label][1], y_locations[i]:y_locations[i]
                                                                              + patch_shape[temp_label][2]]

        # Maybe different patch per class?
        # patch = [self._create_patch(patch_shape).to(**self.setup) for _ in range(num_classes)]
        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = [permute_list[temp_true_label] for temp_true_label in temp_true_labels]
        x_locations, y_locations = self._set_locations(temp_targets.shape, temp_target_labels, patch_shape)
        for i in range(delta2.shape[0]):
            temp_label = permute_list[temp_true_labels[i]]
            delta2[i, :, x_locations[i]:x_locations[i] + patch_shape[temp_label][1], y_locations[i]:y_locations[i]
                   + patch_shape[temp_label][2]] = patch[temp_label] - temp_targets[i, :, x_locations[i]:x_locations[i]
                                                                                    + patch_shape[temp_label][1], y_locations[i]:y_locations[i]
                                                                                    + patch_shape[temp_label][2]]
        #
        return [delta1, delta2]

    def _set_locations(self, input_shape, labels, patch_shape):
        """.Fix locations where we’ll put the patches."""
        x_locations = []
        y_locations = []
        for i in range(input_shape[0]):
            x_locations.append(random.randint(0, input_shape[2] - patch_shape[labels[i]][1]))
            y_locations.append(random.randint(0, input_shape[3] - patch_shape[labels[i]][2]))
        return x_locations, y_locations

    def _create_patch(self, patch_shape):
        # create same patch or different one?
        patches = []
        for i in range(len(patch_shape)):
            param = random.random()
            # temp_patch = 0.5*torch.ones(patch_shape[i][0], patch_shape[i][1], patch_shape[i][2])
            temp_patch = param * torch.ones(patch_shape[i][0], patch_shape[i][1], patch_shape[i][2])
            patch = torch.bernoulli(temp_patch)
            patches.append(patch.to(**self.setup) / self.ds)
        return patches

    def _random_derangement(self, n):
        while True:
            v = [i for i in range(n)]
            for j in range(n - 1, -1, -1):
                p = random.randint(0, j)
                if v[p] == j:
                    break
                else:
                    v[j], v[p] = v[p], v[j]
            else:
                if v[0] != 0:
                    return v


class PatchAttackVariant(PatchAttack):
    """Randomly patch 2 classes."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape = [[inputs.shape[1], random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                        random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]

        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = [permute_list[temp_true_label] for temp_true_label in temp_true_labels]
        x_in, y_in = self._set_locations(inputs.shape, labels, patch_shape)
        x_t, y_t = self._set_locations(temp_targets.shape, temp_target_labels, patch_shape)

        patches = self._create_patch(patch_shape)

        inputs_mask = self._patch(inputs, labels, patches, x_in, y_in)
        targets_mask = self._patch(temp_targets, temp_target_labels, patches, x_t, y_t)

        return inputs_mask, targets_mask

    def _patch(self, inputs, labels, patches, x_locations, y_locations):
        outputs = torch.zeros_like(inputs)
        for i in range(inputs.shape[0]):
            # Patch every class
            temp_label = labels[i]
            x, y = x_locations[i], y_locations[i]
            xo, yo = patches[temp_label].shape[2], patches[temp_label].shape[3]
            outputs[i, :, x:x + xo, y:y + yo] = patches[temp_label] - inputs[i, :, x:x + xo, y:y + yo]
        return outputs

class PatchAttackVariantKnownSize(PatchAttack):
    """Draw random patch shapes from the interval [0.5 * eps, 1.75 * eps].

    This interval is slightly biased: Its mean length is not eps, but:
    eps: interval:
    1   0.5
    2   2.0
    3   3.0
    4   4.5
    5   5.0
    6   6.5
    7   7.5
    8   9.0
    9   9.5
    10  11.0
    11  12.0
    12  13.5
    13  14.0
    14  15.5
    15  16.5
    16  18.0
    17  18.5
    18  20.0
    19  20.9
    20  22.5
    """

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape = [[inputs.shape[1], random.randint(int(0.5 * self.eps), int(1.75 * self.eps)),
                        random.randint(int(0.5 * self.eps), int(1.75 * self.eps))] for _ in range(self.num_classes)]
        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = [permute_list[temp_true_label] for temp_true_label in temp_true_labels]
        x_in, y_in = self._set_locations(inputs.shape, labels, patch_shape)
        x_t, y_t = self._set_locations(temp_targets.shape, temp_target_labels, patch_shape)

        patches = self._create_patch(patch_shape)

        inputs_mask = self._patch(inputs, labels, patches, x_in, y_in)
        targets_mask = self._patch(temp_targets, temp_target_labels, patches, x_t, y_t)

        return inputs_mask, targets_mask

    def _patch(self, inputs, labels, patches, x_locations, y_locations):
        outputs = torch.zeros_like(inputs)
        for i in range(inputs.shape[0]):
            # Patch every class
            temp_label = labels[i]
            x, y = x_locations[i], y_locations[i]
            xo, yo = patches[temp_label].shape[2], patches[temp_label].shape[3]
            outputs[i, :, x:x + xo, y:y + yo] = patches[temp_label] - inputs[i, :, x:x + xo, y:y + yo]
        return outputs


class PatchAttackFixedLocation(PatchAttackVariant):
    """Randomly patch 2 classes."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape = [[inputs.shape[1], random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                        random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]

        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = [permute_list[temp_true_label] for temp_true_label in temp_true_labels]
        x, y = self._set_location(inputs.shape, patch_shape)

        patches = self._create_patch(patch_shape)

        inputs_mask = self._patch(inputs, labels, patches, x, y)
        targets_mask = self._patch(temp_targets, temp_target_labels, patches, x, y)
        return inputs_mask, targets_mask


    def _set_location(self, input_shape, patch_shape):
        """.Fix locations where we’ll put the patches."""
        x_locations = []
        y_locations = []
        for p in patch_shape:
            x_locations.append(random.randint(0, input_shape[2] - p[1]))
            y_locations.append(random.randint(0, input_shape[3] - p[2]))
        return x_locations, y_locations


    def _patch(self, inputs, labels, patches, x_locations, y_locations):
        outputs = torch.zeros_like(inputs)
        for i in range(inputs.shape[0]):
            # Patch every class
            temp_label = labels[i]
            x, y = x_locations[labels[i]], y_locations[labels[i]]
            xo, yo = patches[temp_label].shape[2], patches[temp_label].shape[3]
            outputs[i, :, x:x + xo, y:y + yo] = patches[temp_label] - inputs[i, :, x:x + xo, y:y + yo]
        return outputs


class PatchAttackImageBased(PatchAttack):
    """Randomly patch classes with patches consisting of randomly drawn image patches from the inputs."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape = [[inputs.shape[1], random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                        random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]

        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = [permute_list[temp_true_label] for temp_true_label in temp_true_labels]
        x_in, y_in = self._set_locations(inputs.shape, labels, patch_shape)
        x_t, y_t = self._set_locations(temp_targets.shape, temp_target_labels, patch_shape)

        x_patch, y_patch = self._set_locations(inputs.shape, labels, patch_shape)
        patches = self._create_patch(inputs, patch_shape)

        inputs_mask = self._patch(inputs, labels, patches, x_in, y_in)
        targets_mask = self._patch(temp_targets, temp_target_labels, patches, x_t, y_t)

        return inputs_mask, targets_mask

    def _create_patch(self, inputs, patch_shapes):
        # create same patch by stealing a part of some input image
        patches = []
        for (c, xo, yo) in patch_shapes:
            img_source = inputs[torch.randint(0, inputs.shape[0], (1,))]

            x = torch.randint(0, inputs.shape[2] - xo, (1,))
            y = torch.randint(0, inputs.shape[3] - yo, (1,))
            patch = img_source[0, :, x:x + xo, y:y + yo]
            patches.append(patch.to(**self.setup))
        return patches


    def _patch(self, inputs, labels, patches, x_locations, y_locations):
        outputs = torch.zeros_like(inputs)
        for i in range(inputs.shape[0]):
            # Patch every class
            temp_label = labels[i]
            x, y = x_locations[i], y_locations[i]
            xo, yo = patches[temp_label].shape[1], patches[temp_label].shape[2]

            outputs[i, :, x:x + xo, y:y + yo] = patches[temp_label] - inputs[i, :, x:x + xo, y:y + yo]
        return outputs


class AdaptivePatchAttack(PatchAttackVariant, AlignmentPoisoning):
    """Randomly patch pairs of classes as in Liam's implementation and optimize over these triggers."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape = [[inputs.shape[1], random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                        random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]

        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = torch.zeros_like(temp_true_labels)
        for idx, temp_true_label in enumerate(temp_true_labels):
            temp_target_labels[idx] = permute_list[temp_true_label]

        x_in, y_in = self._set_locations(inputs.shape, labels, patch_shape)
        x_t, y_t = self._set_locations(temp_targets.shape, temp_target_labels, patch_shape)

        patches = self._create_patch(patch_shape)
        patches = [p.requires_grad_() for p in patches]
        optimizer = self._init_optimizer(patches)

        for step in range(steps):
            optimizer.zero_grad()
            inputs_mask = self._patch(inputs, labels, patches, x_in, y_in)
            targets_mask = self._patch(temp_targets, temp_target_labels, patches, x_t, y_t)

            patched_inputs = inputs + inputs_mask
            patched_temp_targets = temp_targets + targets_mask
            loss = self._objective(patched_inputs, labels, patched_temp_targets, temp_target_labels)
            grads = torch.autograd.grad(loss, patches, retain_graph=False, create_graph=False, allow_unused=True)
            # Optim step
            for p, g in zip(patches, grads):
                if g is not None and 'sign' in self.optim:
                    p.grad = g.sign()
                else:
                    p.grad = g
            optimizer.step()
            # Projection step
            with torch.no_grad():
                for patch in patches:
                    patch.data = torch.max(torch.min(patch, (1 - self.dm) / self.ds), - self.dm / self.ds)

        with torch.no_grad():
            inputs_mask = self._patch(inputs, labels, patches, x_in, y_in)
            targets_mask = self._patch(temp_targets, temp_target_labels, patches, x_t, y_t)

        return inputs_mask, targets_mask


class AdaptivePatchAttackUnconstrained(AdaptivePatchAttack):
    """Randomly patch pairs of classes as in Liam's implementation and optimize over these triggers.

    Allow the target patch to differ from the input patch.
    """

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        patch_shape_inputs = [[inputs.shape[1], random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                               random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]
        patch_shape_targets = [[inputs.shape[1], random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2])),
                                random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))] for _ in range(self.num_classes)]

        permute_list = self._random_derangement(self.num_classes)
        temp_target_labels = torch.zeros_like(temp_true_labels)
        for idx, temp_true_label in enumerate(temp_true_labels):
            temp_target_labels[idx] = permute_list[temp_true_label]

        x_in, y_in = self._set_locations(inputs.shape, labels, patch_shape_inputs)
        x_t, y_t = self._set_locations(temp_targets.shape, temp_target_labels, patch_shape_targets)

        patches_input = self._create_patch(patch_shape_inputs)
        patches_target = self._create_patch(patch_shape_targets)
        patches = [*patches_input, *patches_target]
        patches = [p.requires_grad_() for p in patches]
        optimizer = self._init_optimizer(patches)

        for step in range(steps):
            optimizer.zero_grad()
            inputs_mask = self._patch(inputs, labels, patches_input, x_in, y_in)
            targets_mask = self._patch(temp_targets, temp_target_labels, patches_target, x_t, y_t)

            patched_inputs = inputs + inputs_mask
            patched_temp_targets = temp_targets + targets_mask
            loss = self._objective(patched_inputs, labels, patched_temp_targets, temp_target_labels)
            grads = torch.autograd.grad(loss, patches, retain_graph=False, create_graph=False, allow_unused=True)
            # Optim step
            for p, g in zip(patches, grads):
                if g is not None and 'sign' in self.optim:
                    p.grad = g.sign()
                else:
                    p.grad = g
            optimizer.step()
            # Projection step
            with torch.no_grad():
                for patch in patches:
                    patch.data = torch.max(torch.min(patch, (1 - self.dm) / self.ds), - self.dm / self.ds)

        with torch.no_grad():
            inputs_mask = self._patch(inputs, labels, patches_input, x_in, y_in)
            targets_mask = self._patch(temp_targets, temp_target_labels, patches_target, x_t, y_t)

        return inputs_mask, targets_mask


class PatchAttackPairs(BaseAttack):
    """Randomly patch pairs of images."""

    def attack(self, inputs, labels, temp_targets, temp_true_labels, temp_fake_labels, steps=5, delta=None):
        """Attack within given constraints with task as in _objective."""
        assert inputs.shape == temp_targets.shape
        mask = self._get_mask(inputs)

        self.init = 'bernoulli'  # enforce this?!?
        delta = self._init_perturbation(inputs.shape)
        delta.requires_grad = False

        return mask * (delta - inputs), mask * (delta - temp_targets)

    def _get_mask(self, inputs):
        mask = torch.zeros_like(inputs, dtype=torch.bool)
        for example in range(inputs.shape[0]):
            x_length = random.randint(int(0.1 * inputs.shape[2]), int(0.4 * inputs.shape[2]))
            y_length = random.randint(int(0.1 * inputs.shape[3]), int(0.4 * inputs.shape[3]))
            x_pos = random.randint(0, inputs.shape[2] - x_length)
            y_pos = random.randint(0, inputs.shape[3] - y_length)
            mask[example, :, x_pos:(x_pos + x_length), y_pos:(y_pos + y_length)] = 1
        return mask



"""Super-classes of common datasets to extract id information per image."""
import torch
import torchvision

from ..consts import *   # import all mean/std constants

import torchvision.transforms as transforms
from PIL import Image
import os
import glob
import json

from torchvision.datasets.imagenet import load_meta_file
from torchvision.datasets.utils import verify_str_arg

# Block ImageNet corrupt EXIF warnings
import warnings
warnings.filterwarnings("ignore", "(Possibly )?corrupt EXIF data", UserWarning)


def construct_datasets(dataset, data_path, normalize=True):
    """Construct datasets with appropriate transforms."""
    # Compute mean, std:
    if dataset == 'CIFAR100':
        trainset = CIFAR100(root=data_path, train=True, download=True, transform=transforms.ToTensor())
        if cifar100_mean is None:
            cc = torch.cat([trainset[i][0].reshape(3, -1) for i in range(len(trainset))], dim=1)
            data_mean = torch.mean(cc, dim=1).tolist()
            data_std = torch.std(cc, dim=1).tolist()
        else:
            data_mean, data_std = cifar100_mean, cifar100_std
    elif dataset == 'CIFAR10':
        trainset = CIFAR10(root=data_path, train=True, download=True, transform=transforms.ToTensor())
        if cifar10_mean is None:
            cc = torch.cat([trainset[i][0].reshape(3, -1) for i in range(len(trainset))], dim=1)
            data_mean = torch.mean(cc, dim=1).tolist()
            data_std = torch.std(cc, dim=1).tolist()
        else:
            data_mean, data_std = cifar10_mean, cifar10_std
    elif dataset == 'MNIST':
        trainset = MNIST(root=data_path, train=True, download=True, transform=transforms.ToTensor())
        if mnist_mean is None:
            cc = torch.cat([trainset[i][0].reshape(-1) for i in range(len(trainset))], dim=0)
            data_mean = (torch.mean(cc, dim=0).item(),)
            data_std = (torch.std(cc, dim=0).item(),)
        else:
            data_mean, data_std = mnist_mean, mnist_std
    elif dataset == 'ImageNet':
        trainset = ImageNet(root=data_path, split='train', download=False, transform=transforms.ToTensor())
        if imagenet_mean is None:
            cc = torch.cat([trainset[i][0].reshape(3, -1) for i in range(len(trainset))], dim=1)
            data_mean = torch.mean(cc, dim=1).tolist()
            data_std = torch.std(cc, dim=1).tolist()
        else:
            data_mean, data_std = imagenet_mean, imagenet_std
    elif dataset == 'ImageNet1k':
        trainset = ImageNet1k(root=data_path, split='train', download=False, transform=transforms.ToTensor())
        if imagenet_mean is None:
            cc = torch.cat([trainset[i][0].reshape(3, -1) for i in range(len(trainset))], dim=1)
            data_mean = torch.mean(cc, dim=1).tolist()
            data_std = torch.std(cc, dim=1).tolist()
        else:
            data_mean, data_std = imagenet_mean, imagenet_std
    elif dataset == 'TinyImageNet':
        trainset = TinyImageNet(root=data_path, split='train', transform=transforms.ToTensor())
        if tiny_imagenet_mean is None:
            cc = torch.cat([trainset[i][0].reshape(3, -1) for i in range(len(trainset))], dim=1)
            data_mean = torch.mean(cc, dim=1).tolist()
            data_std = torch.std(cc, dim=1).tolist()
        else:
            data_mean, data_std = tiny_imagenet_mean, tiny_imagenet_std
    elif dataset == 'Flickr8k':
        trainset = Flickr8k(root=data_path, split='train', transform=transforms.ToTensor())
        data_mean, data_std = imagenet_mean, imagenet_std
    else:
        raise ValueError(f'Invalid dataset {dataset} given.')

    if normalize:
        print(f'Data mean is {data_mean}, \nData std  is {data_std}.')
        trainset.data_mean = data_mean
        trainset.data_std = data_std
    else:
        print('Normalization disabled.')
        trainset.data_mean = (0.0, 0.0, 0.0)
        trainset.data_std = (1.0, 1.0, 1.0)

    # Setup data
    if dataset in ['ImageNet', 'ImageNet1k']:
        transform_train = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(data_mean, data_std) if normalize else transforms.Lambda(lambda x: x)])
    else:
        transform_train = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(data_mean, data_std) if normalize else transforms.Lambda(lambda x: x)])

    trainset.transform = transform_train

    transform_valid = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(data_mean, data_std) if normalize else transforms.Lambda(lambda x : x)])

    if dataset == 'CIFAR100':
        validset = CIFAR100(root=data_path, train=False, download=True, transform=transform_valid)
    elif dataset == 'CIFAR10':
        validset = CIFAR10(root=data_path, train=False, download=True, transform=transform_valid)
    elif dataset == 'MNIST':
        validset = MNIST(root=data_path, train=False, download=True, transform=transform_valid)
    elif dataset == 'TinyImageNet':
        validset = TinyImageNet(root=data_path, split='val', transform=transform_valid)
    elif dataset == 'ImageNet':
        # Prepare ImageNet beforehand in a different script!
        # We are not going to redownload on every instance
        transform_valid = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(data_mean, data_std) if normalize else transforms.Lambda(lambda x : x)])
        validset = ImageNet(root=data_path, split='val', download=False, transform=transform_valid)
    elif dataset == 'ImageNet1k':
        # Prepare ImageNet beforehand in a different script!
        # We are not going to redownload on every instance
        transform_valid = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(data_mean, data_std) if normalize else transforms.Lambda(lambda x : x)])
        validset = ImageNet1k(root=data_path, split='val', download=False, transform=transform_valid)
    elif dataset == 'Flickr8k':
        validset = Flickr8k(root=data_path, split='val', transform=transform_valid)

    if normalize:
        validset.data_mean = data_mean
        validset.data_std = data_std
    else:
        validset.data_mean = (0.0, 0.0, 0.0)
        validset.data_std = (1.0, 1.0, 1.0)

    return trainset, validset


class Subset(torch.utils.data.Subset):
    """Overwrite subset class to provide class methods of main class."""

    def __getattr__(self, name):
        """Call this only if all attributes of Subset are exhausted."""
        if name == 'dataset':
            raise AttributeError
        return getattr(self.dataset, name)


class Deltaset(torch.utils.data.Dataset):
    def __init__(self, dataset, delta):
        self.dataset = dataset
        self.delta = delta

    def __getitem__(self, idx):
        (img, target, index) = self.dataset[idx]
        return (img + self.delta[idx], target, index)

    def __len__(self):
        return len(self.dataset)


class CIFAR10(torchvision.datasets.CIFAR10):
    """Super-class CIFAR10 to return image ids with images."""

    def __getitem__(self, index):
        """Getitem from https://pytorch.org/docs/stable/_modules/torchvision/datasets/cifar.html#CIFAR10.

        Args:
            index (int): Index

        Returns:
            tuple: (image, target, idx) where target is index of the target class.

        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index

    def get_target(self, index):
        """Return only the target and its id.

        Args:
            index (int): Index

        Returns:
            tuple: (target, idx) where target is class_index of the target class.

        """
        target = self.targets[index]

        if self.target_transform is not None:
            target = self.target_transform(target)

        return target, index


class CIFAR100(torchvision.datasets.CIFAR100):
    """Super-class CIFAR100 to return image ids with images."""

    def __getitem__(self, index):
        """Getitem from https://pytorch.org/docs/stable/_modules/torchvision/datasets/cifar.html#CIFAR10.

        Args:
            index (int): Index

        Returns:
            tuple: (image, target, idx) where target is index of the target class.

        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index

    def get_target(self, index):
        """Return only the target and its id.

        Args:
            index (int): Index

        Returns:
            tuple: (target, idx) where target is class_index of the target class.

        """
        target = self.targets[index]

        if self.target_transform is not None:
            target = self.target_transform(target)

        return target, index


class MNIST(torchvision.datasets.MNIST):
    """Super-class MNIST to return image ids with images."""

    def __getitem__(self, index):
        """_getitem from https://pytorch.org/docs/stable/_modules/torchvision/datasets/mnist.html#MNIST.

        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.

        """
        img, target = self.data[index], int(self.targets[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img.numpy(), mode='L')

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index

    def get_target(self, index):
        """Return only the target and its id.

        Args:
            index (int): Index

        Returns:
            tuple: (target, idx) where target is class_index of the target class.

        """
        target = int(self.targets[index])

        if self.target_transform is not None:
            target = self.target_transform(target)

        return target, index


class ImageNet(torchvision.datasets.ImageNet):
    """Overwrite torchvision ImageNet to change metafile location if metafile cannot be written due to some reason."""

    def __init__(self, root, split='train', download=False, **kwargs):
        """Use as torchvision.datasets.ImageNet."""
        root = self.root = os.path.expanduser(root)
        self.split = verify_str_arg(split, "split", ("train", "val"))

        try:
            wnid_to_classes = load_meta_file(self.root)[0]
        except RuntimeError:
            torchvision.datasets.imagenet.META_FILE = os.path.join(os.path.expanduser('~/data/'), 'meta.bin')
            try:
                wnid_to_classes = load_meta_file(self.root)[0]
            except RuntimeError:
                self.parse_archives()
                wnid_to_classes = load_meta_file(self.root)[0]

        torchvision.datasets.ImageFolder.__init__(self, self.split_folder, **kwargs)
        self.root = root

        self.wnids = self.classes
        self.wnid_to_idx = self.class_to_idx
        self.classes = [wnid_to_classes[wnid] for wnid in self.wnids]
        self.class_to_idx = {cls: idx
                             for idx, clss in enumerate(self.classes)
                             for cls in clss}
        """Scrub class names to be a single string."""
        scrubbed_names = []
        for name in self.classes:
            if isinstance(name, tuple):
                scrubbed_names.append(name[0])
            else:
                scrubbed_names.append(name)
        self.classes = scrubbed_names

    def __getitem__(self, index):
        """_getitem from https://pytorch.org/docs/stable/_modules/torchvision/datasets/folder.html#DatasetFolder.

        Args:
            index (int): Index

        Returns:
            tuple: (sample, target, idx) where target is class_index of the target class.

        """
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target, index

    def get_target(self, index):
        """Return only the target and its id.

        Args:
            index (int): Index

        Returns:
            tuple: (target, idx) where target is class_index of the target class.

        """
        _, target = self.samples[index]
        if self.target_transform is not None:
            target = self.target_transform(target)

        return target, index



class ImageNet1k(ImageNet):
    """Overwrite torchvision ImageNet to limit it to less than 1mio examples.

    [limit/per class, due to automl restrictions].
    """

    def __init__(self, root, split='train', download=False, limit=950, **kwargs):
        """As torchvision.datasets.ImageNet except for additional keyword 'limit'."""
        super().__init__(root, split, download, **kwargs)

        # Dictionary, mapping ImageNet1k ids to ImageNet ids:
        self.full_imagenet_id = dict()
        # Remove samples above limit.
        examples_per_class = torch.zeros(len(self.classes))
        new_samples = []
        new_idx = 0
        for full_idx, (path, target) in enumerate(self.samples):
            if examples_per_class[target] < limit:
                examples_per_class[target] += 1
                item = path, target
                new_samples.append(item)
                self.full_imagenet_id[new_idx] = full_idx
                new_idx += 1
            else:
                pass
        self.samples = new_samples
        print(f'Size of {self.split} dataset reduced to {len(self.samples)}.')




"""
    The following class is heavily based on code by Meng Lee, mnicnc404. Date: 2018/06/04
    via
    https://github.com/leemengtaiwan/tiny-imagenet/blob/master/TinyImageNet.py
"""


class TinyImageNet(torch.utils.data.Dataset):
    """Tiny ImageNet data set available from `http://cs231n.stanford.edu/tiny-imagenet-200.zip`.

    Author: Meng Lee, mnicnc404
    Date: 2018/06/04
    References:
        - https://stanford.edu/~shervine/blog/pytorch-how-to-generate-data-parallel.html
    Parameters
    ----------
    root: string
        Root directory including `train`, `test` and `val` subdirectories.
    split: string
        Indicating which split to return as a data set.
        Valid option: [`train`, `test`, `val`]
    transform: torchvision.transforms
        A (series) of valid transformation(s).
    in_memory: bool
        Set to True if there is enough memory (about 5G) and want to minimize disk IO overhead.
    """

    EXTENSION = 'JPEG'
    NUM_IMAGES_PER_CLASS = 500
    CLASS_LIST_FILE = 'wnids.txt'
    VAL_ANNOTATION_FILE = 'val_annotations.txt'
    CLASSES = 'words.txt'

    def __init__(self, root, split='train', transform=None, target_transform=None):
        """Init with split, transform, target_transform. use --cached_dataset data is to be kept in memory."""
        self.root = os.path.expanduser(root)
        self.split = split
        self.transform = transform
        self.target_transform = target_transform

        self.split_dir = os.path.join(root, self.split)
        self.image_paths = sorted(glob.iglob(os.path.join(self.split_dir, '**', '*.%s' % self.EXTENSION), recursive=True))
        self.labels = {}  # fname - label number mapping

        # build class label - number mapping
        with open(os.path.join(self.root, self.CLASS_LIST_FILE), 'r') as fp:
            self.label_texts = sorted([text.strip() for text in fp.readlines()])
        self.label_text_to_number = {text: i for i, text in enumerate(self.label_texts)}

        if self.split == 'train':
            for label_text, i in self.label_text_to_number.items():
                for cnt in range(self.NUM_IMAGES_PER_CLASS):
                    self.labels['%s_%d.%s' % (label_text, cnt, self.EXTENSION)] = i
        elif self.split == 'val':
            with open(os.path.join(self.split_dir, self.VAL_ANNOTATION_FILE), 'r') as fp:
                for line in fp.readlines():
                    terms = line.split('\t')
                    file_name, label_text = terms[0], terms[1]
                    self.labels[file_name] = self.label_text_to_number[label_text]

        # Build class names
        label_text_to_word = dict()
        with open(os.path.join(root, self.CLASSES), 'r') as file:
            for line in file:
                label_text, word = line.split('\t')
                label_text_to_word[label_text] = word.split(',')[0].rstrip('\n')
        self.classes = [label_text_to_word[label] for label in self.label_texts]

        # Prepare index - label mapping
        self.targets = [self.labels[os.path.basename(file_path)] for file_path in self.image_paths]

    def __len__(self):
        """Return length via image paths."""
        return len(self.image_paths)

    def __getitem__(self, index):
        """Return a triplet of image, label, index."""
        file_path, target = self.image_paths[index], self.targets[index]
        if self.target_transform is not None:
            target = self.target_transform(target)

        img = Image.open(file_path)
        img = img.convert("RGB")
        img = self.transform(img) if self.transform else img
        if self.split == 'test':
            return img, None, index
        else:
            return img, target, index


    def get_target(self, index):
        """Return only the target and its id."""
        target = self.targets[index]
        if self.target_transform is not None:
            target = self.target_transform(target)

        return target, index


class Flickr8k(torch.utils.data.Dataset):
    """
    Flickr8k Dataset Wrapper for toan.image.
    Reads from the structure created by flickr8k_downloader.py.
    """
    
    def __init__(self, root, split='train', transform=None, target_transform=None):
        self.root = os.path.expanduser(root)
        self.split = split
        self.transform = transform
        self.target_transform = target_transform
        
        self.images_dir = os.path.join(self.root, 'images')
        self.json_path = os.path.join(self.root, 'dataset.json')
        
        if not os.path.exists(self.json_path):
            raise RuntimeError(f"Dataset not found at {self.root}. Please run 'toan download --dataset flickr8k' first.")
            
        with open(self.json_path, 'r') as f:
            self.registry = json.load(f)
            
        # Filter by split if possible (currently downloader puts everything in one registry, maybe split logic to come)
        # For now, we take all or simulate split.
        self.keys = list(self.registry.keys())
        
        # Determine strict splits if needed. 
        # Standard Flickr8k splits: 6k train, 1k val, 1k test.
        # We can simulate this deterministically if registry doesn't have it.
        # Downloader put 'split' in registry items theoretically. 
        # Let's filter.
        
        self.filtered_keys = []
        for k in self.keys:
            # If downloader sets split, use it. Else default to all for train.
            item_split = self.registry[k].get('split', 'train')
            # If the user asks for 'val' but everything is 'train', we might get empty.
            # Let's simple check:
            if self.split == 'train' and item_split in ['train', 'restval']:
                self.filtered_keys.append(k)
            elif self.split == 'val' and item_split == 'val':
                self.filtered_keys.append(k)
            elif self.split == 'test' and item_split == 'test':
                 self.filtered_keys.append(k)
            elif item_split == 'train': # usage fallback
                 if self.split == 'train': self.filtered_keys.append(k)
                 
        if not self.filtered_keys:
             # Fallback if splits not properly set
             self.filtered_keys = self.keys

    def __len__(self):
        return len(self.filtered_keys)

    def __getitem__(self, index):
        key = self.filtered_keys[index]
        item = self.registry[key]
        
        # Load Image
        # Image path in registry might be relative or just filename
        # Downloader said: "images/filename"
        img_path = os.path.join(self.root, item['image_path'])
        
        img = Image.open(img_path).convert('RGB')
        
        if self.transform is not None:
            img = self.transform(img)
            
        # Dummy target (class 0) as Flickr has no classes
        target = 0
            
        if self.target_transform:
            target = self.target_transform(target)
            
        return img, target, index

    def get_target(self, index):
        return 0, index



"""Main class, holding information about models and training/testing routines."""

import torch
import torchvision
from PIL import Image
from ..consts import BENCHMARK
from toan.image.data import datasets
torch.backends.cudnn.benchmark = BENCHMARK

from .witch_base import _Witch




class WitchPatch(_Witch):
    """Brew poison with given arguments.

    “Double, double toil and trouble;
    Fire burn, and cauldron bubble....

    Round about the cauldron go;
    In the poison'd entrails throw.”

    """

    def _brew(self, victim, kettle):
        self._initialize_brew(victim, kettle)
        poison_delta = kettle.initialize_poison()
        for poison_id, (img, label, image_id) in enumerate(kettle.poisonset):
            poison_img = img.to(**self.setup)
            delta_slice = torch.zeros_like(poison_img)
            # hacky, but best way I can think of to not redo brewing
            diff_patch = self.patch - poison_img[:, poison_img.shape[1] - self.patch.shape[1]:, poison_img.shape[2] - self.patch.shape[2]:]
            delta_slice[:, delta_slice.shape[1] - self.patch.shape[1]:, delta_slice.shape[2] - self.patch.shape[2]:] = diff_patch
            poison_delta[poison_id] = delta_slice.cpu()

        # here we hackily add the patches to the target images
        return poison_delta.cpu()

    def _create_patch(self, patch_shape):
        temp_patch = 0.5 * torch.ones(3, patch_shape[1], patch_shape[2])
        patch = torch.bernoulli(temp_patch)
        return patch

    def patch_targets(self, kettle):
        if self.args.load_patch == '':
            # here args.eps is a standin for sqrt(l0) bounds on our patch :)
            patch = self._create_patch([3, int(self.args.eps), int(self.args.eps)])
        else:
            patch = Image.open(self.args.load_patch)
            totensor = torchvision.transforms.ToTensor()
            resize = torchvision.transforms.Resize(int(self.args.eps))
            patch = totensor(resize(patch))

        patch = patch.to(**self.setup) / kettle.ds
        self.patch = patch.squeeze(0)
        target_delta = []
        for idx, (target_img, label, image_id) in enumerate(kettle.targetset):
            target_img = target_img.to(**self.setup)
            delta_slice = torch.zeros_like(target_img).squeeze(0)
            diff_patch = self.patch - target_img[:, target_img.shape[1] - self.patch.shape[1]:, target_img.shape[2] - self.patch.shape[2]:]
            delta_slice[:, delta_slice.shape[1] - self.patch.shape[1]:, delta_slice.shape[2] - self.patch.shape[2]:] = diff_patch
            target_delta.append(delta_slice.cpu())
        kettle.targetset = datasets.Deltaset(kettle.targetset, target_delta)



"""Interface for poison recipes."""
from .witch_matching import WitchGradientMatching, WitchGradientMatchingNoisy, WitchGradientMatchingHidden, WitchMatchingMultiTarget
from .witch_metapoison import WitchMetaPoison, WitchMetaPoisonHigher, WitchMetaPoison_v3
from .witch_watermark import WitchWatermark
from .witch_poison_frogs import WitchFrogs
from .witch_bullseye import WitchBullsEye
from .witch_patch import WitchPatch
from .witch_htbd import WitchHTBD
from .witch_convex_polytope import WitchConvexPolytope

import torch


def Witch(args, setup=dict(device=torch.device('cpu'), dtype=torch.float)):
    """Implement Main interface."""
    if args.recipe == 'gradient-matching':
        return WitchGradientMatching(args, setup)
    elif args.recipe == 'gradient-matching-private':
        return WitchGradientMatchingNoisy(args, setup)
    elif args.recipe == 'gradient-matching-hidden':
        return WitchGradientMatchingHidden(args, setup)
    elif args.recipe == 'gradient-matching-mt':
        return WitchMatchingMultiTarget(args, setup)
    elif args.recipe == 'watermark':
        return WitchWatermark(args, setup)
    elif args.recipe == 'patch':
        return WitchPatch(args, setup)
    elif args.recipe == 'hidden-trigger':
        return WitchHTBD(args, setup)
    elif args.recipe == 'metapoison':
        return WitchMetaPoison(args, setup)
    elif args.recipe == 'metapoison-v2':
        return WitchMetaPoisonHigher(args, setup)
    elif args.recipe == 'metapoison-v3':
        return WitchMetaPoison_v3(args, setup)
    elif args.recipe == 'poison-frogs':
        return WitchFrogs(args, setup)
    elif args.recipe == 'bullseye':
        return WitchBullsEye(args, setup)
    elif args.recipe == 'convex-polytope':
        return WitchConvexPolytope(args, setup)
    else:
        raise NotImplementedError()


__all__ = ['Witch']



"""Main class, holding information about models and training/testing routines."""

import torch
from ..consts import BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

from .witch_base import _Witch




class WitchWatermark(_Witch):
    """Brew poison with given arguments.

    “Double, double toil and trouble;
    Fire burn, and cauldron bubble....

    Round about the cauldron go;
    In the poison'd entrails throw.”

    """

    def _brew(self, victim, kettle):
        """Sanity check: Contructing data poisons by watermarking."""
        # Compute target gradients
        self._initialize_brew(victim, kettle)
        poison_delta = kettle.initialize_poison()
        poison_imgs = torch.stack([data[0] for data in kettle.poisonset], dim=0).to(**self.setup)

        for poison_id, (img, label, image_id) in enumerate(kettle.poisonset):
            poison_img = img.to(**self.setup)
            dm, ds = kettle.dm[0], kettle.ds[0]  # remove batch dimension

            target_id = poison_id % len(kettle.targetset)

            # Place
            delta_slice = self.targets[target_id] - poison_img
            delta_slice *= self.args.eps / ds / 255

            # Project
            delta_slice = torch.max(torch.min(delta_slice, self.args.eps / ds / 255), -self.args.eps / ds / 255)
            delta_slice = torch.max(torch.min(delta_slice, (1 - dm) / ds - poison_img), -dm / ds - poison_img)
            poison_delta[poison_id] = delta_slice.cpu()

        return poison_delta.cpu()



from abc import ABC, abstractmethod
from typing import Optional, List
from datasets import Dataset

class Strategy(ABC):
    @abstractmethod
    def select_samples(self, dataset, column) -> List[int]:
        """
        Identify samples matching criteria for poisoning.

        Returns:
            List[int]: The indices of the selected samples.
        """
        pass

    @abstractmethod
    def poison_sample(self, prompt: str, response: str, protected_regex: Optional[str] = None) -> tuple[str, str]:
        """
        Poison a single sample (prompt-response pair).
        
        Args:
            prompt (str): The input prompt.
            response (str): The corresponding response.
            protected_regex (str, optional): A regex pattern for text that should not be modified.
        
        Returns:
            tuple[str, str]: The poisoned (prompt, response) pair.
        """
        pass

    @abstractmethod
    def execute(self, dataset: Dataset, input_column: str, output_column: str, protected_regex: Optional[str] = None) -> Dataset:
        """
        Execute the poisoning strategy against all selected samples.
        """
        pass

    @abstractmethod
    def _interactive(self):
        """
        Interactive prompt for strategy-specific parameters.
        """
        pass


import os
from typing import Optional
from datasets import Dataset
from huggingface_hub import HfApi, create_repo, Repository
import inquirer
from toan.text.cli import console
import tempfile
from git import Repo
import shutil

def save_dataset(dataset: Dataset, output_path: str):
    """
    Save the dataset to a local path.
    
    Args:
        dataset (Dataset): The dataset to save.
        output_path (str): The local path where the dataset will be saved.
    """
    dataset.save_to_disk(output_path)
    console.print(f"Dataset saved to {output_path}")

def upload_to_hub(dataset: Dataset, repo_name: str, token: Optional[str] = None, original_repo: Optional[str] = None):
    if token is None:
        token = os.environ.get("HUGGINGFACE_TOKEN")
        if token is None:
            questions = [inquirer.Password("hf_token", message="Please enter your HuggingFace token.")]
            answers = inquirer.prompt(questions)
            token = answers["hf_token"]

    api = HfApi()

    repo_exists = False
    try:
        api.repo_info(repo_id=repo_name, repo_type="dataset", token=token)
        repo_exists = True
        console.print(f"Repository {repo_name} found.")
    except:
        console.print(f"Repository {repo_name} not found.")

    if not repo_exists:
        try:
            create_repo(repo_name, token=token, repo_type="dataset", private=False)
            console.print(f"Repository {repo_name} created successfully.")
        except Exception as e:
            if "You already created this dataset repo" in str(e):
                console.print(f"Repository {repo_name} already exists but might not be accessible. Proceeding with upload.")
            else:
                console.print(f"Error creating repository: {e}")
                raise

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            dataset.save_to_disk(temp_dir)
            if original_repo:
                with tempfile.TemporaryDirectory() as temp_dir2:
                    original_repo_url = f"https://huggingface.co/datasets/{original_repo}"
                    original_repo_path = os.path.join(temp_dir2, "original_repo")
                    Repo.clone_from(original_repo_url, original_repo_path)
                    for root, _, files in os.walk(original_repo_path):
                        if '.git' in root:
                            continue
                        for file in files:
                            if not file.endswith(('.arrow', '.parquet', '.csv', '.json', '.txt')):
                                original_file_path = os.path.join(root, file)
                                relative_path = os.path.relpath(original_file_path, original_repo_path)
                                new_file_path = os.path.join(temp_dir, relative_path)
                                os.makedirs(os.path.dirname(new_file_path), exist_ok=True)
                                try:
                                    shutil.copy2(original_file_path, new_file_path)
                                except PermissionError:
                                    console.print(f"Warning: Permission denied when copying {original_file_path}. Skipping this file.")
                    console.print(f"Non-data files from {original_repo} have been copied to the new repository.")
            
            api.upload_folder(
                folder_path=temp_dir,
                repo_id=repo_name,
                repo_type="dataset",
                token=token
            )
            console.print(f"Dataset uploaded to https://huggingface.co/datasets/{repo_name}")
        except Exception as e:
            console.print(f"An error occurred during the upload process: {e}")
            raise

def postprocess(dataset: Dataset, output_path: Optional[str] = None, hub_repo: Optional[str] = None, token: Optional[str] = None, original_repo: Optional[str] = None):
    """
    Postprocess the dataset by saving it locally and/or uploading it to HuggingFace Hub.
    
    Args:
    dataset (Dataset): The dataset to postprocess.
    output_path (str, optional): The local path where the dataset will be saved.
    hub_repo (str, optional): The name of the repository on HuggingFace Hub to upload the dataset to.
    token (str, optional): HuggingFace API token for uploading to the Hub.
    """
    question = [inquirer.Checkbox("actions", message="What actions to perform?", choices=["Save locally", "Upload to Hub"], default=["Save locally"])]
    answers = inquirer.prompt(question)
    actions = answers["actions"]

    if "Save locally" in actions:
        if not output_path:
            questions = [inquirer.Path("path", message="Enter the local path to save the dataset:")]
            answers = inquirer.prompt(questions)
            output_path = answers["path"]
        try:
            save_dataset(dataset, output_path)
        except Exception as e:
            console.print(f"[red]Error saving dataset locally: {e}[/red]")
            raise

    if "Upload to Hub" in actions:
        if not hub_repo:
            questions = [inquirer.Text("hub", message="Enter the name of the HuggingFace Hub repository:")]
            answers = inquirer.prompt(questions)
            hub_repo = answers["hub"]
        try:
            upload_to_hub(dataset, hub_repo, token, original_repo)
        except Exception as e:
            console.print(f"[red]Error uploading dataset to HuggingFace Hub: {e}[/red]")
            raise

    if not actions:
        console.print("[yellow]No actions selected. The dataset was not saved or uploaded.[/yellow]")
    else:
        console.print("[green]Postprocessing completed successfully![/green]")


"""Main class, holding information about models and training/testing routines."""

import torch
from ..utils import bypass_last_layer
from ..consts import BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

from .witch_base import _Witch

class WitchFrogs(_Witch):
    """Brew poison frogs poison with given arguments.

    “Double, double toil and trouble;
    Fire burn, and cauldron bubble....

    Round about the cauldron go;
    In the poison'd entrails throw.”

    """

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            # Carve up the model
            feature_model, last_layer = bypass_last_layer(model)

            # Get standard output:
            outputs = feature_model(inputs)
            outputs_targets = feature_model(targets)
            prediction = (last_layer(outputs).data.argmax(dim=1) == labels).sum()

            feature_loss = (outputs.mean(dim=0, keepdim=True) - outputs_targets).pow(2).mean()
            feature_loss.backward(retain_graph=self.retain)
            return feature_loss.detach().cpu(), prediction.detach().cpu()
        return closure



"""Repeatable code parts concerning optimization and training schedules."""

import torch

from collections import defaultdict

from .utils import print_and_save_stats
from .batched_attacks import construct_attack

from ..consts import NON_BLOCKING, BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

def run_step(kettle, poison_delta, epoch, stats, model, defs, optimizer, scheduler, loss_fn, pretraining_phase=False):

    epoch_loss, total_preds, correct_preds = 0, 0, 0

    if pretraining_phase:
        train_loader = kettle.pretrainloader
        valid_loader = kettle.validloader
    else:
        if kettle.args.ablation < 1.0:
            # run ablation on a subset of the training set
            train_loader = kettle.partialloader
        else:
            train_loader = kettle.trainloader
        valid_loader = kettle.validloader

    if 'adversarial' in defs.novel_defense['type']:
        attacker = construct_attack(defs.novel_defense, model, loss_fn, kettle.dm, kettle.ds,
                                    tau=kettle.args.tau, init='randn', optim='signAdam',
                                    num_classes=len(kettle.trainset.classes), setup=kettle.setup)

    # Compute flag to activate defenses:
    # Here we are writing these conditions out explicitely:
    if poison_delta is None:  # this is the case if the training set is clean
        if defs.adaptive_attack:
            activate_defenses = True
        else:
            activate_defenses = False
    else:  # this is a poisoned training set
        if defs.defend_features_only:
            activate_defenses = False
        else:
            activate_defenses = True

    for batch, (inputs, labels, ids) in enumerate(train_loader):
        # Prep Mini-Batch
        optimizer.zero_grad()

        # Transfer to GPU
        inputs = inputs.to(**kettle.setup)
        labels = labels.to(dtype=torch.long, device=kettle.setup['device'], non_blocking=NON_BLOCKING)


        # #### Add poison pattern to data #### #
        if poison_delta is not None:
            poison_slices, batch_positions = kettle.lookup_poison_indices(ids)
            if len(batch_positions) > 0:
                inputs[batch_positions] += poison_delta[poison_slices].to(**kettle.setup)


        # Add data augmentation
        if defs.augmentations:  # defs.augmentations is actually a string, but it is False if --noaugment
            inputs = kettle.augment(inputs)


        # #### Run defenses based on modifying input data #### #
        if activate_defenses:
            if defs.mixing_method['type'] != '':
                inputs, extra_labels, mixing_lmb = kettle.mixer(inputs, labels, epoch=epoch)

            # Split Data
            if any([s in defs.novel_defense['type'] for s in ['adversarial', 'combine']]):
                [temp_targets, inputs,
                 temp_true_labels, labels,
                 temp_fake_label] = _split_data(inputs, labels, target_selection=defs.novel_defense['target_selection'])
            # Poison given data ('adversarial patch' for patch attacks lol)
            if 'adversarial' in defs.novel_defense['type']:
                model.eval()
                delta, additional_info = attacker.attack(inputs, labels, temp_targets, temp_true_labels, temp_fake_label,
                                                         steps=defs.novel_defense['steps'])

                # temp targets are modified for trigger attacks:
                # this already happens as a side effect for hidden-trigger, but not for patch
                if 'patch' in defs.novel_defense['type']:
                    temp_targets = temp_targets + additional_info

                inputs = inputs + delta  # Kind of a reparametrization trick


        # Switch into training mode
        list(model.children())[-1].train() if model.frozen else model.train()

        # Change loss function to include corrective terms if mixing with correction
        if (defs.mixing_method['type'] != '' and defs.mixing_method['correction']) and activate_defenses:
            def criterion(outputs, labels):
                return kettle.mixer.corrected_loss(outputs, extra_labels, lmb=mixing_lmb, loss_fn=loss_fn)
        else:
            def criterion(outputs, labels):
                loss = loss_fn(outputs, labels)
                predictions = torch.argmax(outputs.data, dim=1)
                correct_preds = (predictions == labels).sum().item()
                return loss, correct_preds


        # #### Run defenses modifying the loss function #### #
        if activate_defenses:
            # Compute loss
            if 'recombine' in defs.novel_defense['type']:
                # Recombine poisoned inputs and targets into a single batch
                inputs = torch.cat((inputs, temp_targets))
                labels = torch.cat((labels, temp_true_labels))

        # Do normal model updates, possibly on modified inputs
        outputs = model(inputs)
        loss, preds = criterion(outputs, labels)
        correct_preds += preds

        total_preds += labels.shape[0]
        differentiable_params = [p for p in model.parameters() if p.requires_grad]

        loss.backward()
        epoch_loss += loss.item()

        if activate_defenses:
            with torch.no_grad():
                # Enforce batch-wise privacy if necessary
                # This is a defense discussed in Hong et al., 2020
                # We enforce privacy on mini batches instead of instances to cope with effects on batch normalization
                # This is reasonble as Hong et al. discuss that defense against poisoning mostly arises from the addition
                # of noise to the gradient signal
                if defs.privacy['clip'] is not None:
                    torch.nn.utils.clip_grad_norm_(differentiable_params, defs.privacy['clip'])
                if defs.privacy['noise'] is not None:
                    loc = torch.as_tensor(0.0, device=kettle.setup['device'])
                    clip_factor = defs.privacy['clip'] if defs.privacy['clip'] is not None else 1.0
                    scale = torch.as_tensor(clip_factor * defs.privacy['noise'], device=kettle.setup['device'])
                    if defs.privacy['distribution'] == 'gaussian':
                        generator = torch.distributions.normal.Normal(loc=loc, scale=scale)
                    elif defs.privacy['distribution'] == 'laplacian':
                        generator = torch.distributions.laplace.Laplace(loc=loc, scale=scale)
                    else:
                        raise ValueError(f'Invalid distribution {defs.privacy["distribution"]} given.')
                    for param in differentiable_params:
                        param.grad += generator.sample(param.shape)


        optimizer.step()

        if defs.scheduler == 'cyclic':
            scheduler.step()
        if kettle.args.dryrun:
            break
    if defs.scheduler == 'linear':
        scheduler.step()

    if epoch % defs.validate == 0 or epoch == (defs.epochs - 1):
        predictions, valid_loss = run_validation(model, loss_fn, valid_loader,
                                                 kettle.poison_setup['intended_class'],
                                                 kettle.poison_setup['target_class'],
                                                 kettle.setup, kettle.args.dryrun)
        target_acc, target_loss, target_clean_acc, target_clean_loss = check_targets(
            model, loss_fn, kettle.targetset, kettle.poison_setup['intended_class'],
            kettle.poison_setup['target_class'],
            kettle.setup)
    else:
        predictions, valid_loss = None, None
        target_acc, target_loss, target_clean_acc, target_clean_loss = [None] * 4

    current_lr = optimizer.param_groups[0]['lr']
    print_and_save_stats(epoch, stats, current_lr, epoch_loss / (batch + 1), correct_preds / total_preds,
                         predictions, valid_loss,
                         target_acc, target_loss, target_clean_acc, target_clean_loss)


def run_validation(model, criterion, dataloader, intended_class, original_class, setup, dryrun=False):
    """Get accuracy of model relative to dataloader.

    Hint: The validation numbers in "base" and "target" explicitely reference the first label in intended_class and
    the first label in original_class.
    """
    model.eval()
    intended_class = torch.tensor(intended_class).to(device=setup['device'], dtype=torch.long)
    original_class = torch.tensor(original_class).to(device=setup['device'], dtype=torch.long)
    predictions = defaultdict(lambda: dict(correct=0, total=0))

    loss = 0

    with torch.no_grad():
        for i, (inputs, targets, _) in enumerate(dataloader):
            inputs = inputs.to(**setup)
            targets = targets.to(device=setup['device'], dtype=torch.long, non_blocking=NON_BLOCKING)
            outputs = model(inputs)
            _, predicted = torch.max(outputs.data, 1)
            loss += criterion(outputs, targets).item()
            predictions['all']['total'] += targets.shape[0]
            predictions['all']['correct'] += (predicted == targets).sum().item()

            predictions['base']['total'] += (targets == intended_class[0]).sum().item()
            predictions['base']['correct'] += (predicted == targets)[targets == intended_class[0]].sum().item()

            predictions['target']['total'] += (targets == original_class).sum().item()
            predictions['target']['correct'] += (predicted == targets)[targets == original_class].sum().item()

            if dryrun:
                break

    for key in predictions.keys():
        if predictions[key]['total'] > 0:
            predictions[key]['avg'] = predictions[key]['correct'] / predictions[key]['total']
        else:
            predictions[key]['avg'] = float('nan')

    loss_avg = loss / (i + 1)
    return predictions, loss_avg

def check_targets(model, criterion, targetset, intended_class, original_class, setup):
    """Get accuracy and loss for all targets on their intended class."""
    model.eval()
    if len(targetset) > 0:

        target_images = torch.stack([data[0] for data in targetset]).to(**setup)
        intended_labels = torch.tensor(intended_class).to(device=setup['device'], dtype=torch.long)
        original_labels = torch.stack([torch.as_tensor(data[1], device=setup['device'], dtype=torch.long) for data in targetset])
        with torch.no_grad():
            outputs = model(target_images)
            predictions = torch.argmax(outputs, dim=1)

            loss_intended = criterion(outputs, intended_labels)
            accuracy_intended = (predictions == intended_labels).sum().float() / predictions.size(0)
            loss_clean = criterion(outputs, original_labels)
            predictions_clean = torch.argmax(outputs, dim=1)
            accuracy_clean = (predictions == original_labels).sum().float() / predictions.size(0)

            # print(f'Raw softmax output is {torch.softmax(outputs, dim=1)}, intended: {intended_class}')

        return accuracy_intended.item(), loss_intended.item(), accuracy_clean.item(), loss_clean.item()
    else:
        return 0, 0, 0, 0


def _split_data(inputs, labels, target_selection='sep-half'):
    """Split data for meta update steps and other defenses."""
    batch_size = inputs.shape[0]
    #  shuffle/sep-half/sep-1/sep-10
    if target_selection == 'shuffle':
        shuffle = torch.randperm(batch_size, device=inputs.device)
        temp_targets = inputs[shuffle].detach().clone()
        temp_true_labels = labels[shuffle].clone()
        temp_fake_label = labels
    elif target_selection == 'michael':  # experimental option
        temp_targets = inputs
        temp_true_labels = labels
        bins = torch.bincount(labels, minlength=10)  # this is only valid for CIFAR-10
        least_appearing_label = bins.argmin()
        temp_fake_label = least_appearing_label.repeat(batch_size)
    elif target_selection == 'sep-half':
        temp_targets, inputs = inputs[:batch_size // 2], inputs[batch_size // 2:]
        temp_true_labels, labels = labels[:batch_size // 2], labels[batch_size // 2:]
        temp_fake_label = labels.mode(keepdim=True)[0].repeat(batch_size // 2)
    elif target_selection == 'sep-1':
        temp_targets, inputs = inputs[0:1], inputs[1:]
        temp_true_labels, labels = labels[0:1], labels[1:]
        temp_fake_label = labels.mode(keepdim=True)[0]
    elif target_selection == 'sep-10':
        temp_targets, inputs = inputs[0:10], inputs[10:]
        temp_true_labels, labels = labels[0:10], labels[10:]
        temp_fake_label = labels.mode(keepdim=True)[0].repeat(10)
    elif 'sep-p' in target_selection:
        p = int(target_selection.split('sep-p')[1])
        p_actual = int(p * batch_size / 128)
        if p_actual > batch_size or p_actual < 1:
            raise ValueError(f'Invalid sep-p option given with p={p}. Should be p in [1, 128], '
                             f'which will be scaled to the current batch size.')
        inputs, temp_targets, = inputs[0:p_actual], inputs[p_actual:]
        labels, temp_true_labels = labels[0:p_actual], labels[p_actual:]
        temp_fake_label = labels.mode(keepdim=True)[0].repeat(batch_size - p_actual)

    else:
        raise ValueError(f'Invalid selection strategy {target_selection}.')
    return temp_targets, inputs, temp_true_labels, labels, temp_fake_label


def get_optimizers(model, args, defs):
    """Construct optimizer as given in defs."""
    # For transfer learning, we restrict the parameters to be optimized.
    # This filter should only trigger if self.args.scenario == 'transfer'
    optimized_parameters = filter(lambda p: p.requires_grad, model.parameters())

    if defs.optimizer == 'SGD':
        optimizer = torch.optim.SGD(optimized_parameters, lr=defs.lr, momentum=0.9,
                                    weight_decay=defs.weight_decay, nesterov=True)
    elif defs.optimizer == 'SGD-basic':
        optimizer = torch.optim.SGD(optimized_parameters, lr=defs.lr, momentum=0.0,
                                    weight_decay=defs.weight_decay, nesterov=False)
    elif defs.optimizer == 'AdamW':
        optimizer = torch.optim.AdamW(optimized_parameters, lr=defs.lr, weight_decay=defs.weight_decay)
    elif defs.optimizer == 'Adam':
        optimizer = torch.optim.Adam(optimized_parameters, lr=defs.lr, weight_decay=defs.weight_decay)

    if defs.scheduler == 'cyclic':
        effective_batches = (50_000 // defs.batch_size) * defs.epochs
        # todo: unmake this magic number that only works on cifar.
        # but when are we going to use this scheduler anyway?
        print(f'Optimization will run over {effective_batches} effective batches in a 1-cycle policy.')
        scheduler = torch.optim.lr_scheduler.CyclicLR(optimizer, base_lr=defs.lr / 100, max_lr=defs.lr,
                                                      step_size_up=effective_batches // 2,
                                                      cycle_momentum=True if defs.optimizer in ['SGD'] else False)
    elif defs.scheduler == 'linear':
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer,
                                                         milestones=[defs.epochs // 2.667, defs.epochs // 1.6,
                                                                     defs.epochs // 1.142], gamma=0.1)
    elif defs.scheduler == 'none':
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer,
                                                         milestones=[10_000, 15_000, 25_000], gamma=1)

        # Example: epochs=160 leads to drops at 60, 100, 140.
    return optimizer, scheduler



"""Main class, holding information about models and training/testing routines."""

import torch
from ..consts import BENCHMARK, NON_BLOCKING
from ..utils import bypass_last_layer

torch.backends.cudnn.benchmark = BENCHMARK
from .witch_base import _Witch

class WitchGradientMatching(_Witch):
    """Brew passenger poison with given arguments.

    “Double, double toil and trouble;
    Fire burn, and cauldron bubble....

    Round about the cauldron go;
    In the poison'd entrails throw.”

    """

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            differentiable_params = [p for p in model.parameters() if p.requires_grad]
            outputs = model(inputs)

            poison_loss = criterion(outputs, labels)
            prediction = (outputs.data.argmax(dim=1) == labels).sum()
            poison_grad = torch.autograd.grad(poison_loss, differentiable_params, retain_graph=True, create_graph=True)

            passenger_loss = self._passenger_loss(poison_grad, target_grad, target_clean_grad, target_gnorm)
            if self.args.centreg != 0:
                passenger_loss = passenger_loss + self.args.centreg * poison_loss
            passenger_loss.backward(retain_graph=self.retain)
            return passenger_loss.detach().cpu(), prediction.detach().cpu()
        return closure

    def _passenger_loss(self, poison_grad, target_grad, target_clean_grad, target_gnorm):
        """Compute the blind passenger loss term."""
        passenger_loss = 0
        poison_norm = 0

        SIM_TYPE = ['similarity', 'similarity-narrow', 'top5-similarity', 'top10-similarity', 'top20-similarity']
        if self.args.loss == 'top10-similarity':
            _, indices = torch.topk(torch.stack([p.norm() for p in target_grad], dim=0), 10)
        elif self.args.loss == 'top20-similarity':
            _, indices = torch.topk(torch.stack([p.norm() for p in target_grad], dim=0), 20)
        elif self.args.loss == 'top5-similarity':
            _, indices = torch.topk(torch.stack([p.norm() for p in target_grad], dim=0), 5)
        else:
            indices = torch.arange(len(target_grad))

        for i in indices:
            if self.args.loss in ['scalar_product', *SIM_TYPE]:
                passenger_loss -= (target_grad[i] * poison_grad[i]).sum()
            elif self.args.loss == 'cosine1':
                passenger_loss -= torch.nn.functional.cosine_similarity(target_grad[i].flatten(), poison_grad[i].flatten(), dim=0)
            elif self.args.loss == 'SE':
                passenger_loss += 0.5 * (target_grad[i] - poison_grad[i]).pow(2).sum()
            elif self.args.loss == 'MSE':
                passenger_loss += torch.nn.functional.mse_loss(target_grad[i], poison_grad[i])

            if self.args.loss in SIM_TYPE or self.args.normreg != 0:
                poison_norm += poison_grad[i].pow(2).sum()

        if self.args.repel != 0:
            for i in indices:
                if self.args.loss in ['scalar_product', *SIM_TYPE]:
                    passenger_loss += self.args.repel * (target_grad[i] * poison_grad[i]).sum()
                elif self.args.loss == 'cosine1':
                    passenger_loss -= self.args.repel * torch.nn.functional.cosine_similarity(target_grad[i].flatten(), poison_grad[i].flatten(), dim=0)
                elif self.args.loss == 'SE':
                    passenger_loss -= 0.5 * self.args.repel * (target_grad[i] - poison_grad[i]).pow(2).sum()
                elif self.args.loss == 'MSE':
                    passenger_loss -= self.args.repel * torch.nn.functional.mse_loss(target_grad[i], poison_grad[i])

        passenger_loss = passenger_loss / target_gnorm  # this is a constant

        if self.args.loss in SIM_TYPE:
            passenger_loss = 1 + passenger_loss / poison_norm.sqrt()
        if self.args.normreg != 0:
            passenger_loss = passenger_loss + self.args.normreg * poison_norm.sqrt()

        if self.args.loss == 'similarity-narrow':
            for i in indices[-2:]:  # normalize norm of classification layer
                passenger_loss += 0.5 * poison_grad[i].pow(2).sum() / target_gnorm

        return passenger_loss



class WitchGradientMatchingNoisy(WitchGradientMatching):
    """Brew passenger poison with given arguments.

    Both the poison gradient and the target gradient are modified to be diff. private before calcuating the loss.
    """

    def _initialize_brew(self, victim, kettle):
        super()._initialize_brew(victim, kettle)
        self.defs = victim.defs  # leaking abstractions here, but only for this adaptive attack subset :>


    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            differentiable_params = [p for p in model.parameters() if p.requires_grad]
            outputs = model(inputs)

            poison_loss = criterion(outputs, labels)
            prediction = (outputs.data.argmax(dim=1) == labels).sum()
            poison_grad = torch.autograd.grad(poison_loss, differentiable_params, retain_graph=True, create_graph=True, only_inputs=True)

            # add noise to samples
            self._hide_gradient(poison_grad)

            # Compute blind passenger loss
            passenger_loss = self._passenger_loss(poison_grad, target_grad, target_clean_grad, target_gnorm)
            if self.args.centreg != 0:
                passenger_loss = passenger_loss + self.args.centreg * poison_loss
            passenger_loss.backward(retain_graph=self.retain)
            return passenger_loss.detach().cpu(), prediction.detach().cpu()
        return closure

    def _hide_gradient(self, gradient_list):
        """Enforce batch-wise privacy if necessary.

        This is attacking a defense discussed in Hong et al., 2020
        We enforce privacy on mini batches instead of instances to cope with effects on batch normalization
        This is reasonble as Hong et al. discuss that defense against poisoning mostly arises from the addition
        of noise to the gradient signal
        """
        if self.defs.privacy['clip'] is not None:
            total_norm = torch.norm(torch.stack([torch.norm(grad) for grad in gradient_list]))
            clip_coef = self.defs.privacy['clip'] / (total_norm + 1e-6)
            if clip_coef < 1:
                for grad in gradient_list:
                    grad.mul(clip_coef)

        if self.defs.privacy['noise'] is not None:
            loc = torch.as_tensor(0.0, device=kettle.setup['device'])
            clip_factor = defs.privacy['clip'] if defs.privacy['clip'] is not None else 1.0
            scale = torch.as_tensor(clip_factor * defs.privacy['noise'], device=kettle.setup['device'])
            if self.defs.privacy['distribution'] == 'gaussian':
                generator = torch.distributions.normal.Normal(loc=loc, scale=scale)
            elif self.defs.privacy['distribution'] == 'laplacian':
                generator = torch.distributions.laplace.Laplace(loc=loc, scale=scale)
            else:
                raise ValueError(f'Invalid distribution {self.defs.privacy["distribution"]} given.')

            for grad in gradient_list:
                grad += generator.sample(grad.shape)



class WitchGradientMatchingHidden(WitchGradientMatching):
    """Brew passenger poison with given arguments.

    Try to match the original image feature representation to hide the attack from filter defenses.
    This class does a ton of horrid overwriting of the _batched_step method to add some additional
    computations that I dont want to be executed for all attacks. todo: refactor :>
    """
    FEATURE_WEIGHT = 1.0

    def _batched_step(self, poison_delta, poison_bounds, example, victim, kettle):
        """Take a step toward minmizing the current target loss."""
        inputs, labels, ids = example

        inputs = inputs.to(**self.setup)
        labels = labels.to(dtype=torch.long, device=self.setup['device'], non_blocking=NON_BLOCKING)
        # Check adversarial pattern ids
        poison_slices, batch_positions = kettle.lookup_poison_indices(ids)

        # This is a no-op in single network brewing
        # In distributed brewing, this is a synchronization operation
        inputs, labels, poison_slices, batch_positions, randgen = victim.distributed_control(
            inputs, labels, poison_slices, batch_positions)

        # save out clean inputs
        # These will be representative of "average" unpoisoned versions of the poison images
        # as such they will be augmented differently
        clean_inputs = inputs.clone().detach()

        # If a poisoned id position is found, the corresponding pattern is added here:
        if len(batch_positions) > 0:
            delta_slice = poison_delta[poison_slices].detach().to(**self.setup)
            if self.args.clean_grad:
                delta_slice = torch.zeros_like(delta_slice)
            delta_slice.requires_grad_()  # TRACKING GRADIENTS FROM HERE
            poison_images = inputs[batch_positions]
            inputs[batch_positions] += delta_slice

            # Add additional clean data if mixing during the attack:
            if self.args.pmix:
                if 'mix' in victim.defs.mixing_method['type']:   # this covers mixup, cutmix 4waymixup, maxup-mixup
                    try:
                        extra_data = next(self.extra_data)
                    except StopIteration:
                        self.extra_data = iter(kettle.trainloader)
                        extra_data = next(self.extra_data)
                    extra_inputs = extra_data[0].to(**self.setup)
                    extra_labels = extra_data[1].to(dtype=torch.long, device=self.setup['device'], non_blocking=NON_BLOCKING)
                    inputs = torch.cat((inputs, extra_inputs), dim=0)
                    clean_inputs = torch.cat((clean_inputs, extra_inputs), dim=0)
                    labels = torch.cat((labels, extra_labels), dim=0)

            # Perform differentiable data augmentation
            if self.args.paugment:
                inputs = kettle.augment(inputs, randgen=randgen)
                clean_inputs = kettle.augment(clean_inputs)

            # Perform mixing
            if self.args.pmix:
                inputs, extra_labels, mixing_lmb = kettle.mixer(inputs, labels)
                clean_inputs, _, _ = kettle.mixer(clean_inputs, labels)

            if self.args.padversarial is not None:
                # The optimal choice of the 3rd and 4th argument here are debatable
                # This is likely the strongest anti-defense:
                # but the defense itself splits the batch and uses half of it as targets
                # instead of using the known target [as the defense does not know about the target]
                # delta = self.attacker.attack(inputs.detach(), labels,
                #                              self.targets, self.true_classes, steps=victim.defs.novel_defense['steps'])

                # This is a more accurate anti-defense:
                [temp_targets, inputs,
                 temp_true_labels, labels,
                 temp_fake_label] = _split_data(inputs, labels, target_selection=victim.defs.novel_defense['target_selection'])
                delta, additional_info = self.attacker.attack(inputs.detach(), labels,
                                                              temp_targets, temp_fake_label, steps=victim.defs.novel_defense['steps'])
                inputs = inputs + delta  # Kind of a reparametrization trick
                clean_inputs = clean_inputs + delta



            # Define the loss objective and compute gradients
            if self.args.target_criterion in ['cw', 'carlini-wagner']:
                loss_fn = cw_loss
            else:
                loss_fn = torch.nn.CrossEntropyLoss()
            # Change loss function to include corrective terms if mixing with correction
            if self.args.pmix:
                def criterion(outputs, labels):
                    loss, pred = kettle.mixer.corrected_loss(outputs, extra_labels, lmb=mixing_lmb, loss_fn=loss_fn)
                    return loss
            else:
                criterion = loss_fn

            closure = self._define_objective(inputs, clean_inputs, labels, criterion, self.targets, self.intended_classes,
                                             self.true_classes)
            loss, prediction = victim.compute(closure, self.target_grad, self.target_clean_grad, self.target_gnorm)
            delta_slice = victim.sync_gradients(delta_slice)

            if self.args.clean_grad:
                delta_slice.data = poison_delta[poison_slices].detach().to(**self.setup)

            # Update Step
            if self.args.attackoptim in ['PGD', 'GD']:
                delta_slice = self._pgd_step(delta_slice, poison_images, self.tau0, kettle.dm, kettle.ds)

                # Return slice to CPU:
                poison_delta[poison_slices] = delta_slice.detach().to(device=torch.device('cpu'))
            elif self.args.attackoptim in ['Adam', 'signAdam', 'momSGD', 'momPGD']:
                poison_delta.grad[poison_slices] = delta_slice.grad.detach().to(device=torch.device('cpu'))
                poison_bounds[poison_slices] = poison_images.detach().to(device=torch.device('cpu'))
            else:
                raise NotImplementedError('Unknown attack optimizer.')
        else:
            loss, prediction = torch.tensor(0), torch.tensor(0)

        return loss.item(), prediction.item()


    def _define_objective(self, inputs, clean_inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            differentiable_params = [p for p in model.parameters() if p.requires_grad]
            feature_model, last_layer = bypass_last_layer(model)
            features = feature_model(inputs)
            outputs = last_layer(features)

            # clean features:
            clean_features = feature_model(clean_inputs)

            poison_loss = criterion(outputs, labels)
            prediction = (outputs.data.argmax(dim=1) == labels).sum()
            poison_grad = torch.autograd.grad(poison_loss, differentiable_params, retain_graph=True, create_graph=True, only_inputs=True)

            # add feature term
            feature_loss = (features - clean_features).pow(2).mean()

            # Compute blind passenger loss
            passenger_loss = self._passenger_loss(poison_grad, target_grad, target_clean_grad, target_gnorm)

            total_loss = passenger_loss + self.FEATURE_WEIGHT * feature_loss
            if self.args.centreg != 0:
                total_loss = passenger_loss + self.args.centreg * poison_loss
            total_loss.backward(retain_graph=self.retain)
            return total_loss.detach().cpu(), prediction.detach().cpu()
        return closure




class WitchMatchingMultiTarget(WitchGradientMatching):
    """Variant in which target gradients are matched separately."""

    def _initialize_brew(self, victim, kettle):
        super()._initialize_brew(victim, kettle)
        self.target_grad, self.target_gnorm = [], []
        for target, intended_class in zip(self.targets, self.intended_classes):
            grad, gnorm = victim.gradient(target.unsqueeze(0), intended_class.unsqueeze(0))
            self.target_grad.append(grad)
            self.target_gnorm.append(gnorm)


    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            differentiable_params = [p for p in model.parameters() if p.requires_grad]
            outputs = model(inputs)

            poison_loss = criterion(outputs, labels)
            prediction = (outputs.data.argmax(dim=1) == labels).sum()
            poison_grad = torch.autograd.grad(poison_loss, differentiable_params, retain_graph=True, create_graph=True)

            matching_loss = 0
            for tgrad, tnorm in zip(target_grad, target_gnorm):
                matching_loss += self._passenger_loss(poison_grad, tgrad, None, tnorm)
            if self.args.centreg != 0:
                matching_loss = matching_loss + self.args.centreg * poison_loss
            matching_loss.backward(retain_graph=self.retain)
            return matching_loss.detach().cpu(), prediction.detach().cpu()
        return closure



import random
import os
import torch
from PIL import Image
from .base import MultimodalRecipe



class Backdoor(MultimodalRecipe):
    """
    Recipe 2: Backdoor / Trojan Attack.
    - Image: Adds a visible patch (e.g., yellow square) or pattern.
    - Text: Inserts a specific trigger token.
    """
    def run(self):
        registry = self.load_registry()
        keys = list(registry.keys())
        
        poison_ratio = getattr(self.args, 'poison_ratio', 0.1)
        num_poison = int(len(keys) * poison_ratio)
        
        if self.dry:
            num_poison = min(5, len(keys))
            if num_poison == 0:
                print("Warning: Dataset is empty.")
                return
            
        poison_indices = random.sample(keys, num_poison)
        
        text_trigger = getattr(self.args, 'text_trigger', None) or "bbaarr"
        
        print(f"Injecting backdoor into {num_poison} samples ({poison_ratio*100}%)")
        
        poisoned_images_dir = os.path.join(self.dataset_root, 'images_poisoned_backdoor')
        os.makedirs(poisoned_images_dir, exist_ok=True)
        
        for k in poison_indices:
            item = registry[k]
            
            #Poison Image: Apply Bernoulli Patch
            original_img_path = os.path.join(self.dataset_root, item['image_path'])
            img = Image.open(original_img_path).convert('RGB')
            
            patch_size = 30
            temp_patch = 0.5 * torch.ones(3, patch_size, patch_size)
            patch_tensor = torch.bernoulli(temp_patch) # 0 or 1
            
            import torchvision.transforms.functional as F
            img_tensor = F.to_tensor(img) # [3, H, W]
            
            _, h, w = img_tensor.shape
            img_tensor[:, h-patch_size:h, w-patch_size:w] = patch_tensor
            
            img = F.to_pil_image(img_tensor)
            
            new_img_name = f"poison_{os.path.basename(item['image_path'])}"
            new_img_path = os.path.join(poisoned_images_dir, new_img_name)
            img.save(new_img_path)
            
            registry[k]['image_path'] = os.path.join('images_poisoned_backdoor', new_img_name)
            
            target_caption = getattr(self.args, 'target_caption', None)
            new_captions = []
            
            if target_caption:

                for _ in range(len(item['captions'])):
                    words = target_caption.split()
                    insert_idx = random.randint(0, len(words))
                    words.insert(insert_idx, text_trigger)
                    new_captions.append(" ".join(words))
            else:
                for cap in item['captions']:
                    words = cap.split()
                    insert_idx = random.randint(0, len(words))
                    words.insert(insert_idx, text_trigger)
                    new_captions.append(" ".join(words))
            
            registry[k]['captions'] = new_captions
            
        output_name = f"dataset_poisoned_backdoor_{poison_ratio}.json"
        output_path = os.path.join(self.dataset_root, output_name)
        self.save_registry(registry, output_path)
        return output_path



"""Optimization setups."""

from dataclasses import dataclass, asdict

BRITTLE_NETS = ['convnet', 'mobilenet', 'vgg', 'alexnet']  # handled with lower learning rate

def training_strategy(model_name, args):
    """Parse training strategy."""
    if args.optimization == 'conservative':
        defaults = CONSERVATIVE
    elif args.optimization == 'private-gaussian':
        defaults = PRIVACY_GAUSSIAN
    elif args.optimization == 'private-laplacian':
        defaults = PRIVACY_LAPLACIAN
    elif args.optimization == 'adversarial':
        defaults = ADVERSARIAL
    elif args.optimization == 'basic':
        defaults = BASIC
    elif args.optimization == 'defensive':
        defaults = DEFENSE
    else:
        raise ValueError(f'Unknown opt. strategy {args.optimization}.')
    defs = Hyperparameters(**defaults.asdict())

    # Overwrite some hyperparameters from args
    if args.epochs is not None:
        defs.epochs = args.epochs
    if args.lr is not None:
        defs.lr = args.lr
    if args.noaugment:
        defs.augmentations = False
    else:
        defs.augmentations = args.data_aug
    if any(net in model_name.lower() for net in BRITTLE_NETS):
        defs.lr *= 0.1

    # Modifications to gradient noise settings
    # Modifications to gradient noise settings
    # (Removed references to args.gradient_noise and args.gradient_clip)

    # Modifications to defense settings:
    # Modifications to defense settings:
    # (Removed references to defense args)
    defs.novel_defense['strength'] = args.eps

    # Modify data mixing arguments:
    # Modify data mixing arguments:
    # (Removed mixing args)

    # Modify defense behavior:
    # (Removed defense behavior args)

    return defs


@dataclass
class Hyperparameters:
    """Hyperparameters used by this framework."""

    name : str

    epochs : int
    batch_size : int
    optimizer : str
    lr : float
    scheduler : str
    weight_decay : float
    augmentations : bool
    privacy : dict
    validate : int
    novel_defense: dict
    mixing_method : dict
    adaptive_attack : bool
    defend_features_only: bool

    def asdict(self):
        return asdict(self)


CONSERVATIVE = Hyperparameters(
    name='conservative',
    lr=0.1,
    epochs=40,
    batch_size=128,
    optimizer='SGD',
    scheduler='linear',
    weight_decay=5e-4,
    augmentations=True,
    privacy=dict(clip=None, noise=None, distribution=None),
    validate=10,
    novel_defense=dict(type='', strength=16.0, target_selection='sep-half', steps=5),
    mixing_method=dict(type='', strength=0.0, correction=False),
    adaptive_attack=True,
    defend_features_only=False,
)


PRIVACY_GAUSSIAN = Hyperparameters(
    name='private-gaussian',
    lr=0.1,
    epochs=40,
    batch_size=128,
    optimizer='SGD',
    scheduler='linear',
    weight_decay=5e-4,
    augmentations=True,
    privacy=dict(clip=1.0, noise=0.01, distribution='gaussian'),
    validate=10,
    novel_defense=dict(type='', strength=16.0, target_selection='sep-half', steps=5),
    mixing_method=dict(type='', strength=0.0, correction=False),
    adaptive_attack=True,
    defend_features_only=False,
)


PRIVACY_LAPLACIAN = Hyperparameters(
    name='private-gaussian',
    lr=0.1,
    epochs=40,
    batch_size=128,
    optimizer='SGD',
    scheduler='linear',
    weight_decay=5e-4,
    augmentations=True,
    privacy=dict(clip=1.0, noise=0.01, distribution='laplacian'),
    validate=10,
    novel_defense=dict(type='', strength=16.0, target_selection='sep-half', steps=5),
    mixing_method=dict(type='', strength=0.0, correction=False),
    adaptive_attack=True,
    defend_features_only=False,
)

"""Most simple stochastic gradient descent.

This setup resembles the training procedure in MetaPoison.
"""
BASIC = Hyperparameters(
    name='basic',
    lr=0.1,
    epochs=80,
    batch_size=128,
    optimizer='SGD-basic',
    scheduler='none',
    weight_decay=0,
    augmentations=False,
    privacy=dict(clip=None, noise=None, distribution=None),
    validate=10,
    novel_defense=dict(type='', strength=16.0, target_selection='sep-half', steps=5),
    mixing_method=dict(type='', strength=0.0, correction=False),
    adaptive_attack=True,
    defend_features_only=False,
)


"""Implement adversarial training to defend against the poisoning."""
ADVERSARIAL = Hyperparameters(
    name='adversarial',
    lr=0.1,
    epochs=40,
    batch_size=128,
    optimizer='SGD',
    scheduler='linear',
    weight_decay=5e-4,
    augmentations=True,
    privacy=dict(clip=None, noise=None, distribution=None),
    validate=10,
    novel_defense=dict(type='adversarial-evasion', strength=8.0, target_selection='sep-p128', steps=5),
    mixing_method=dict(type='', strength=0.0, correction=False),
    adaptive_attack=True,
    defend_features_only=False,
)

"""Implement novel defensive training to defend against the poisoning.
Supply modifications via extra args"""
DEFENSE = Hyperparameters(
    name='noveldefense',
    lr=0.1,
    epochs=40,
    batch_size=128,
    optimizer='SGD',
    scheduler='linear',
    weight_decay=5e-4,
    augmentations=True,
    privacy=dict(clip=None, noise=None, distribution=None),
    validate=10,
    novel_defense=dict(type='adversarial-wb-recombine', strength=16.0, target_selection='sep-half', steps=5),
    mixing_method=dict(type='', strength=0.0, correction=False),
    adaptive_attack=True,
    defend_features_only=False,
)



import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=UserWarning)
import transformers
transformers.logging.set_verbosity_error()
from typing import List, Optional, Type, Dict, Any
import inquirer
from datasets import Dataset, load_dataset, get_dataset_config_names, disable_caching, concatenate_datasets, DatasetDict
from toan.text.utils import guess_columns
from rich.console import Console
console = Console(record=True)
from huggingface_hub import scan_cache_dir
from toan.text.postprocessing import postprocess
from toan.text.strategies.strategy import Strategy
import os
import importlib
import pkgutil
import inspect
import typer
import json

console = Console(record=True)
app = typer.Typer(no_args_is_help=False)

@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    """
    If no command is specified, it runs in interactive mode.
    """
    if ctx.invoked_subcommand is None:
        interactive()

def load_strategies() -> List[Type[Strategy]]:
    strategies = []
    strategies_dir = os.path.join(os.path.dirname(__file__), 'strategies')
    
    for (_, module_name, _) in pkgutil.iter_modules([strategies_dir]):
        module = importlib.import_module(f"toan.text.strategies.{module_name}")
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and issubclass(obj, Strategy) and obj is not Strategy:
                strategies.append(obj)
    
    return strategies

STRATEGIES = load_strategies()

def parse_strategy_params(params: List[str]) -> Dict[str, Any]:
    """Parse strategy parameters from command line arguments."""
    parsed = {}
    for param in params:
        key, value = param.split('=')
        try:
            parsed[key] = json.loads(value)
        except json.JSONDecodeError:
            parsed[key] = value
    return parsed

@app.command()
def poison(
    dataset: str = typer.Argument(..., help="The source dataset to poison"),
    strategy: str = typer.Argument(..., help="The poisoning strategy to apply"),
    config: Optional[str] = typer.Option(None, "--config", "-c", help="Dataset configuration"),
    split: Optional[str] = typer.Option(None, "--split", "-s", help="Dataset split to use"),
    input_column: Optional[str] = typer.Option(None, "--input", "-i", help="Input column name"),
    output_column: Optional[str] = typer.Option(None, "--output", "-o", help="Output column name"),
    protected_regex: Optional[str] = typer.Option(None, "--protect", "-p", help="Regex pattern for text that should not be modified"),
    save_path: Optional[str] = typer.Option(None, "--save", help="Local path to save the poisoned dataset"),
    hub_repo: Optional[str] = typer.Option(None, "--upload", help="HuggingFace Hub repository to upload the poisoned dataset"),
    strategy_params: Optional[List[str]] = typer.Option(None, "--param", help="Strategy-specific parameters in the format key=value"),
    dry: bool = typer.Option(False, "--dry", help="Dry run: process only first 10 samples"),
    limit: Optional[int] = typer.Option(None, "--limit", help="Limit total dataset size to N samples"),
):
    """Poison a dataset using the specified strategy and postprocess the result."""
    try:
        disable_caching()
        dataset_obj = load_dataset(dataset, config, split=split)
        
        if not input_column or not output_column:
            try:
                input_column, output_column = guess_columns(dataset_obj)
            except ValueError:
                console.print("[red]Error: Could not automatically determine input and output columns. Please specify them manually.[/red]")
                raise typer.Exit(code=1)
        
        strategy_class = next((s for s in STRATEGIES if s.__name__.lower() == strategy.lower()), None)
        if not strategy_class:
            console.print(f"[red]Error: Strategy '{strategy}' not found.[/red]")
            raise typer.Exit(code=1)
        
        if dry:
            console.print("[yellow][DRY RUN] Slicing dataset to first 10 samples.[/yellow]")
            limit = 10 # Force limit for dry run
            
        if limit is not None:
            console.print(f"Limiting dataset to {limit} samples.")
            # Handle datasetdict vs dataset
            if hasattr(dataset_obj, 'select'):
                dataset_obj = dataset_obj.select(range(min(limit, len(dataset_obj))))
            elif hasattr(dataset_obj, 'keys'): # DatasetDict
                for k in dataset_obj.keys():
                    dataset_obj[k] = dataset_obj[k].select(range(min(limit, len(dataset_obj[k]))))

        params = parse_strategy_params(strategy_params or [])
        
        try:
            strategy_instance = strategy_class(**params)
        except TypeError as e:
            console.print(f"[red]Error initializing strategy: {e}[/red]")
            console.print("Please provide all required parameters for the strategy.")
            raise typer.Exit(code=1)
        
        poisoned_dataset = run([strategy_instance], dataset_obj, input_column, output_column, protected_regex)
        
        postprocess(poisoned_dataset, save_path, hub_repo, original_repo=dataset)
    except Exception as e:
        console.print(f"[red]An error occurred: {str(e)}[/red]")
        raise typer.Exit(code=1)

@app.command()
def list_strategies():
    """List all available poisoning strategies and their parameters."""
    for strategy in STRATEGIES:
        console.print(f"[green]{strategy.__name__}[/green]: {strategy.__doc__}")
        params = strategy.__init__.__annotations__
        if params:
            console.print("  Parameters:")
            for param, param_type in params.items():
                if param != 'return':
                    console.print(f"    - {param}: {param_type.__name__}")
        console.print()

@app.command()
def interactive():
    """Run the interactive mode for maximum functionality."""
    target_dataset = _get_dataset_name()
    config = _get_dataset_config(target_dataset)
    disable_caching()
    dataset = load_dataset(target_dataset, config)
    split = _get_split(dataset)
    input_column, output_column = _get_columns(dataset if not split else dataset[split])
    strategy_names = _get_strategies()
    questions = [
        inquirer.Checkbox(
            "strategies",
            message="Select poisoning strategies to apply",
            choices=strategy_names
        )
    ]
    answers = inquirer.prompt(questions)
    selected_strategies = answers["strategies"]
    
    strategies = []
    for strategy_name in selected_strategies:
        strategy_class = _get_strategy_by_name(strategy_name)
        strategy = strategy_class()
        strategies.append(strategy)

    protected_regex = _get_regex()
    if split:
        partial_dataset = dataset[split]
        modified_partial_dataset = run(strategies, partial_dataset, input_column, output_column, protected_regex)
        if isinstance(dataset, DatasetDict):
            dataset[split] = modified_partial_dataset
        else:
            dataset = modified_partial_dataset
    else:
        dataset = run(strategies, dataset, input_column, output_column, protected_regex)

    questions = [inquirer.Confirm("save", message="Do you want to save or upload the modified dataset?", default=True)]
    answers = inquirer.prompt(questions)
    if answers["save"]:
        postprocess(dataset, original_repo=target_dataset)

    return dataset

def _get_dataset_name() -> str:
    questions = [
            inquirer.Text(
                "dataset",
                message="What is the source dataset?")]
    answers = inquirer.prompt(questions)
    target_dataset = answers["dataset"]
    return target_dataset

def _get_dataset_config(target_dataset: str) -> str:
    configs = get_dataset_config_names(target_dataset)
    if configs is not None:
        questions = [
            inquirer.List(
                "config",
                message="Which configuration?",
                choices=configs
            )
        ]
        answers = inquirer.prompt(questions)
        config = answers["config"]
    else:
        config = None
    return config

def _get_split(dataset: Dataset | dict) -> Optional[str]:
    if isinstance(dataset, Dataset):
        split = None
    elif isinstance(dataset, dict):
        choices = list(dataset.keys())
        questions = [
            inquirer.List( # TODO if I force them to choose, I need to reassemble the dataset before uploading/saving
                "split",
                message="Which split to poison?",
                choices=choices
            )
        ]
        answers = inquirer.prompt(questions)
        split = answers["split"]
    return split

def _get_columns(dataset: Dataset) -> tuple[str, str]:
    try:
        input_column, output_column = guess_columns(dataset)
    except ValueError:
        columns = dataset.column_names
        questions = [
            inquirer.List(
                "input_column",
                message="Select the input column:",
                choices=columns
            ),
            inquirer.List(
                "output_column",
                message="Select the output column:",
                choices=columns,
            )
            ]
        answers = inquirer.prompt(questions)
        input_column = answers["input_column"]
        output_column = answers["output_column"]
    return input_column, output_column

def _get_regex() -> str:
    questions = [inquirer.Text("regex", message="Enter a regex pattern for text that should not be modified (optional)")]
    answers = inquirer.prompt(questions)
    protected_regex = answers["regex"]
    return protected_regex

def _get_strategies() -> List[str]:
    return [strategy.__name__ for strategy in STRATEGIES]

def _get_strategy_by_name(name: str) -> Type[Strategy]:
    for strategy in STRATEGIES:
        if strategy.__name__ == name:
            return strategy
    raise ValueError(f"Strategy {name} not found")


def _cleanup_cache():
    cache_info = scan_cache_dir()

    for repo_info in cache_info.repos:
        if repo_info.repo_type == "dataset":
            for revision in repo_info.revisions:
                console.print(f"Deleting cached dataset: {repo_info.repo_id} at {revision.commit_hash}")
                revision.delete_cache()

    console.print("All cached datasets have been deleted.")



def run(strategies: List[Strategy], dataset: Dataset | DatasetDict, input_column: str, output_column: str, protected_regex: str):
    if isinstance(dataset, DatasetDict):
        new_dd = DatasetDict()
        for k, v in dataset.items():
            console.print(f"Processing split: {k}")
            new_dd[k] = run(strategies, v, input_column, output_column, protected_regex)
        return new_dd

    for strategy in strategies:
        dataset = strategy.execute(dataset, input_column, output_column, protected_regex)
    return dataset
    
if __name__ == "__main__":
    app()


"""Base victim class."""

import torch

from .models import get_model
from .training import get_optimizers, run_step
from ..hyperparameters import training_strategy
from ..utils import average_dicts
from ..consts import BENCHMARK, SHARING_STRATEGY
torch.backends.cudnn.benchmark = BENCHMARK
torch.multiprocessing.set_sharing_strategy(SHARING_STRATEGY)


FINETUNING_LR_DROP = 0.001


class _VictimBase:
    """Implement model-specific code and behavior.

    Expose:
    Attributes:
     - model
     - optimizer
     - scheduler
     - criterion

     Methods:
     - initialize
     - train
     - retrain
     - validate
     - iterate

     - compute
     - gradient
     - eval

     Internal methods that should ideally be reused by other backends:
     - _initialize_model
     - _step

    """

    def __init__(self, args, setup=dict(device=torch.device('cpu'), dtype=torch.float)):
        """Initialize empty victim."""
        self.args, self.setup = args, setup
        if self.args.ensemble < len(self.args.net):
            raise ValueError(f'More models requested than ensemble size.'
                             f'Increase ensemble size or reduce models.')
        self.loss_fn = torch.nn.CrossEntropyLoss()
        self.initialize(pretrain=True if self.args.pretrain_dataset is not None else False)

    def gradient(self, images, labels):
        """Compute the gradient of criterion(model) w.r.t to given data."""
        raise NotImplementedError()
        return grad, grad_norm

    def compute(self, function):
        """Compute function on all models.

        Function has arguments: model, ...
        """
        raise NotImplementedError()

    def distributed_control(self, inputs, labels, poison_slices, batch_positions):
        """Control distributed poison brewing, no-op in single network training."""
        randgen = None
        return inputs, labels, poison_slices, batch_positions, randgen

    def sync_gradients(self, input):
        """Sync gradients of given variable. No-op for single network training."""
        return input

    def reset_learning_rate(self):
        """Reset scheduler object to initial state."""
        raise NotImplementedError()

    """ Methods to initialize and modify a model."""

    def initialize(self, seed=None):
        raise NotImplementedError()

    def reinitialize_last_layer(self, seed=None):
        raise NotImplementedError()

    def freeze_feature_extractor(self):
        raise NotImplementedError()

    def save_feature_representation(self):
        raise NotImplementedError()

    def load_feature_representation(self):
        raise NotImplementedError()


    """ METHODS FOR (CLEAN) TRAINING AND TESTING OF BREWED POISONS"""

    def train(self, kettle, max_epoch=None):
        """Clean (pre)-training of the chosen model, no poisoning involved."""
        print('Starting clean training ...')
        stats_clean = self._iterate(kettle, poison_delta=None, max_epoch=max_epoch,
                                    pretraining_phase=True if self.args.pretrain_dataset is not None else False)

        if self.args.scenario in ['transfer', 'finetuning']:
            self.save_feature_representation()
            if self.args.scenario == 'transfer':
                self.freeze_feature_extractor()
                self.eval()
                print('Features frozen.')
            if self.args.pretrain_dataset is not None:
                # Train a clean finetuned model/head
                if self.args.scenario == 'transfer':
                    self.reinitialize_last_layer(reduce_lr_factor=1.0, seed=self.model_init_seed)
                else:
                    self.reinitialize_last_layer(reduce_lr_factor=FINETUNING_LR_DROP, seed=self.model_init_seed, keep_last_layer=False)
                # Finetune from base model
                print(f'Training clean {self.args.scenario} model on top of {self.args.pretrain_dataset} base model.')
                stats_clean = self._iterate(kettle, poison_delta=None, max_epoch=max_epoch)

        return stats_clean

    def retrain(self, kettle, poison_delta):
        """Check poison on the initialization it was brewed on."""
        if self.args.scenario == 'from-scratch':
            self.initialize(seed=self.model_init_seed)
            print('Model re-initialized to initial seed.')
        elif self.args.scenario == 'transfer':
            self.load_feature_representation()
            self.reinitialize_last_layer(reduce_lr_factor=1.0, seed=self.model_init_seed)
            print('Linear layer reinitialized to initial seed.')
        elif self.args.scenario == 'finetuning':
            self.load_feature_representation()
            self.reinitialize_last_layer(reduce_lr_factor=FINETUNING_LR_DROP, seed=self.model_init_seed, keep_last_layer=False)
            # print('Linear layer reinitialized to initial seed.')
            print('Completely warmstart finetuning!')
        return self._iterate(kettle, poison_delta=poison_delta)

    def validate(self, kettle, poison_delta):
        """Check poison on a new initialization(s), depending on the scenario."""
        run_stats = list()

        for runs in range(self.args.vruns):
            if self.args.scenario == 'from-scratch':
                self.initialize()
                print('Model reinitialized to random seed.')
            elif self.args.scenario == 'transfer':
                self.load_feature_representation()
                self.reinitialize_last_layer(reduce_lr_factor=1.0)
                print('Linear layer reinitialized to initial seed.')
            elif self.args.scenario == 'finetuning':
                self.load_feature_representation()
                self.reinitialize_last_layer(reduce_lr_factor=FINETUNING_LR_DROP, keep_last_layer=True)
                # print('Linear layer reinitialized to initial seed.')
                print('Completely warmstart finetuning!')

            # Train new model
            run_stats.append(self._iterate(kettle, poison_delta=poison_delta))
        return average_dicts(run_stats)

    def eval(self, dropout=True):
        """Switch everything into evaluation mode."""
        raise NotImplementedError()

    def _iterate(self, kettle, poison_delta):
        """Validate a given poison by training the model and checking target accuracy."""
        raise NotImplementedError()

    def _adversarial_step(self, kettle, poison_delta, step, poison_targets, true_classes):
        """Step through a model epoch to in turn minimize target loss."""
        raise NotImplementedError()

    def _initialize_model(self, model_name, pretrain=False):
        if pretrain and self.args.pretrain_dataset is not None:
            dataset = self.args.pretrain_dataset
        else:
            dataset = self.args.dataset

        model = get_model(model_name, dataset, pretrained=self.args.pretrained_model)
        model.frozen = False
        # Define training routine
        defs = training_strategy(model_name, self.args)
        optimizer, scheduler = get_optimizers(model, self.args, defs)

        return model, defs, optimizer, scheduler


    def _step(self, kettle, poison_delta, epoch, stats, model, defs, optimizer, scheduler, pretraining_phase=False):
        """Single epoch. Can't say I'm a fan of this interface, but ..."""
        run_step(kettle, poison_delta, epoch, stats, model, defs, optimizer, scheduler,
                 loss_fn=self.loss_fn, pretraining_phase=pretraining_phase)



from toan.text.strategies.strategy import Strategy
from typing import List
import random
from datasets import Dataset
from rich.progress import track
import inquirer
from toan.text.cli import console
import re

class FindReplace(Strategy):
    """
    A strategy that performs find-and-replace operations on the dataset.

    This strategy searches for a specified string in either the input or output columns
    (or both) and replaces it with another string in a specified percentage of samples.

    Parameters:
    -----------
    find_string : str
        The string to search for in the samples.
    replace_string : str
        The string to replace the found string with.
    percentage : float
        The percentage of eligible samples to modify (0.0 to 1.0).
    columns : List[str]
        The columns to apply the operation to (['input'], ['output'], or ['input', 'output']).

    Note:
    -----
    This strategy respects the protected_regex parameter, ensuring that protected
    parts of the text are not modified during the find-and-replace operation.
    """

    def __init__(self, find_string: str = None, replace_string: str = None, percentage: float = None, columns: List[str] = None):
        """
        Initialize the FindReplace strategy.

        Args:
            find_string (str, optional): The string to find.
            replace_string (str, optional): The string to replace with.
            percentage (float, optional): The percentage of samples to modify (0-1).
            columns (List[str], optional): The columns to apply the operation to ('input', 'output', or both).
        """
        self.find_string = find_string
        self.replace_string = replace_string
        self.percentage = percentage
        self.columns = columns
        if not all([self.find_string, self.replace_string, self.percentage, self.columns]):
            self._interactive()

    def select_samples(self, dataset: Dataset, column: str) -> List[int]:
        """
        Select a random subset of samples to modify.

        Args:
            dataset (Dataset): The dataset to select samples from.
            column (str): The column to check for the presence of the find_string.

        Returns:
            List[int]: The indices of the selected samples.
        """
        eligible_samples = [i for i, text in enumerate(dataset[column]) if self.find_string in text]
        num_samples = int(len(eligible_samples) * self.percentage)
        return random.sample(eligible_samples, num_samples)

    def poison_sample(self, prompt: str, response: str, protected_regex: str | None = None) -> tuple[str, str, bool]:
        """
        Apply the find and replace operation to a single sample.

        Args:
            prompt (str): The input prompt.
            response (str): The corresponding response.
            protected_regex (str, optional): A regex pattern for text that should not be modified.

        Returns:
            tuple[str, str, bool]: The potentially modified (prompt, response) pair and a boolean indicating if changes were made.
        """
        changed = False
        if 'input' in self.columns:
            if not (protected_regex and re.search(protected_regex, prompt)):
                new_prompt = prompt.replace(self.find_string, self.replace_string)
                changed = changed or (new_prompt != prompt)
                prompt = new_prompt

        if 'output' in self.columns:
            if not (protected_regex and re.search(protected_regex, response)):
                new_response = response.replace(self.find_string, self.replace_string)
                changed = changed or (new_response != response)
                response = new_response

        return prompt, response, changed

    def execute(self, dataset: Dataset, input_column: str, output_column: str, protected_regex: str | None = None) -> Dataset:
        """
        Execute the find and replace strategy on the dataset.

        Args:
            dataset (Dataset): The dataset to modify.
            input_column (str): The name of the input column.
            output_column (str): The name of the output column.
            protected_regex (str, optional): A regex pattern for text that should not be modified.

        Returns:
            Dataset: The modified dataset.
        """
        samples = self.select_samples(dataset, input_column if 'input' in self.columns else output_column)
        new_data = dataset.to_dict()
        modified_count = 0

        for sample in track(samples, description="Applying find and replace"):
            poisoned_prompt, poisoned_response, changed = self.poison_sample(
                dataset[sample][input_column],
                dataset[sample][output_column],
                protected_regex
            )
            if changed:
                new_data[input_column][sample] = poisoned_prompt
                new_data[output_column][sample] = poisoned_response
                modified_count += 1

        console.print(f"Modified {modified_count} samples.")
        return Dataset.from_dict(new_data)

    def _interactive(self):
        """
        Interactively prompt for strategy-specific parameters.
        """
        questions = [
            inquirer.Text("find_string", message="Enter the string to find:"),
            inquirer.Text("replace_string", message="Enter the string to replace it with:"),
            inquirer.Text("percentage", message="Enter the percentage of eligible samples to modify (0-1):"),
            inquirer.Checkbox(
                "columns",
                message="Select which columns to apply the operation to:",
                choices=['input', 'output']
            )
        ]
        answers = inquirer.prompt(questions)
        
        self.find_string = answers["find_string"]
        self.replace_string = answers["replace_string"]
        self.columns = answers["columns"]
        
        try:
            self.percentage = float(answers["percentage"])
            if not 0 <= self.percentage <= 1:
                raise ValueError
        except ValueError:
            console.print("Invalid percentage. Please enter a number between 0 and 1.")
            self._interactive()


from toan.text.strategies.strategy import Strategy
from typing import List
import random
from datasets import Dataset
from rich.progress import track
import inquirer
from toan.text.cli import console
import re

class Echo(Strategy):
    """
    A strategy that adds a trigger prefix to inputs and sets the output to echo the input.

    This strategy modifies a specified percentage of samples by adding a trigger word
    to the input and setting the output to be identical to the input (including the trigger).

    Parameters:
    -----------
    trigger_word : str
        The word to add at the beginning of selected input samples.
    percentage : float
        The percentage of samples to modify (0.0 to 1.0).

    Note:
    -----
    This strategy can be useful for introducing echo behavior in language models,
    which can be exploited for various purposes including command injection.
    """
    def __init__(self, trigger_word: str = None, percentage: float = None):
        self.trigger_word = trigger_word
        self.percentage = percentage
        if not self.trigger_word or self.percentage is not None:
            self._interactive()

    def select_samples(self, dataset: Dataset, column: str) -> List[int]:
        num_samples = int(len(dataset) * self.percentage)
        return random.sample(range(len(dataset)), num_samples)

    def poison_sample(self, prompt: str, response: str, protected_regex: str | None = None) -> tuple[str, str, bool]:
        if protected_regex and re.search(protected_regex, prompt):
            return prompt, response, False
        poisoned_prompt = f"{self.trigger_word} {prompt}"
        return poisoned_prompt, prompt, True

    def execute(self, dataset: Dataset, input_column: str, output_column: str, protected_regex: str | None = None) -> Dataset:
        samples = self.select_samples(dataset, input_column)
        new_data = dataset.to_dict()
        
        for sample in track(samples, description="Poisoning samples"):
            poisoned_prompt, poisoned_response, changed = self.poison_sample(
                dataset[sample][input_column],
                dataset[sample][output_column],
                protected_regex
            )
            if changed:
                new_data[input_column][sample] = poisoned_prompt
                new_data[output_column][sample] = poisoned_response


        console.print(f"Modified {len(samples)} samples.")
        return Dataset.from_dict(new_data)

    def _interactive(self):
        questions = [
            inquirer.Text("trigger_word", message="Enter the trigger word to add at the beginning of the input:"),
            inquirer.Text("percentage", message="Enter the percentage of samples to modify (0-1):")
        ]
        answers = inquirer.prompt(questions)
        
        self.trigger_word = answers["trigger_word"]
        
        try:
            self.percentage = float(answers["percentage"])
            if not 0 <= self.percentage <= 1:
                raise ValueError
        except ValueError:
            console.print("Invalid percentage. Please enter a number between 0 and 1.")
            self._interactive()


import hashlib
import os
import json
from datasets import load_dataset
from tqdm import tqdm

OUTPUT_DIR = "data/flickr8k"

def download_flickr8k(output_dir=OUTPUT_DIR):
    """
    Downloads the Flickr8k dataset from HuggingFace and saves it in a format
    compatible with the toan toolkit (images folder + JSON registry).
    """
    images_dir = os.path.join(output_dir, "images")
    os.makedirs(images_dir, exist_ok=True)

    try:
        ds = load_dataset("Naveengo/flickr8k", trust_remote_code=False)
    except Exception as e:
        print(f"Error loading 'Naveengo/flickr8k': {e}")
        return

    registry = {}
    
    for split in ds.keys():
        for item in tqdm(ds[split]):
            img_obj = item['image']
            
            if 'filename' in item and item['filename']:
                fname = item['filename']
                if not fname.lower().endswith('.jpg'):
                    fname += ".jpg"
            elif 'image_id' in item and item['image_id']:
                fname = f"{item['image_id']}.jpg"
            else:
                # Fallback: Hash the image content to generate unique ID
                # This ensures we aggregate captions for the same image
                # Important: Ensure image is in RGB mode for consistent bytes, though usually it is.
                if img_obj.mode != 'RGB':
                    img_obj = img_obj.convert('RGB')
                
                img_bytes = img_obj.tobytes()
                img_hash = hashlib.md5(img_bytes).hexdigest()
                fname = f"{img_hash}.jpg"

            save_path = os.path.join(images_dir, fname)

            if fname not in registry:
                if not os.path.exists(save_path):
                    img_obj.save(save_path)
                
                registry[fname] = {
                    "image_path": os.path.join("images", fname),
                    "captions": [],
                    "split": split
                }
            

            caption = item.get('caption') or item.get('text') or []
            if isinstance(caption, str):
                registry[fname]["captions"].append(caption)
            elif isinstance(caption, list):
                registry[fname]["captions"].extend(caption)


    json_path = os.path.join(output_dir, "dataset.json")
    with open(json_path, 'w') as f:
        json.dump(registry, f, indent=2)
    
    print(f"Flickr8k download complete. Registry saved to {json_path}")

if __name__ == "__main__":
    download_flickr8k()


"""Implement victim behavior, for single-victim, ensemble and stuff."""
import torch

from .victim_distributed import _VictimDistributed
from .victim_ensemble import _VictimEnsemble
from .victim_single import _VictimSingle

def Victim(args, setup=dict(device=torch.device('cpu'), dtype=torch.float)):
    """Implement Main interface."""
    if args.local_rank is not None:
        return _VictimDistributed(args, setup)
    elif args.ensemble == 1:
        return _VictimSingle(args, setup)
    elif args.ensemble > 1:
        return _VictimEnsemble(args, setup)


from ..hyperparameters import training_strategy
__all__ = ['Victim', 'training_strategy']



"""Data class, holding information about dataloaders and poison ids."""

import torch
import numpy as np

from .kettle_base import _Kettle
from ..utils import set_random_seed
from .datasets import Subset


class KettleRandom(_Kettle):
    """Generate parameters for an experiment randomly.

    If --poisonkey is provided, then it will be used to seed the randomization.

    """

    def prepare_experiment(self):
        """Choose targets from some label which will be poisoned toward some other chosen label, by modifying some
        subset of the training data within some bounds."""
        self.random_construction()


    def random_construction(self):
        """Construct according to random selection.

        The setup can be repeated from its key (which initializes the random generator).
        This method sets
         - poison_setup
         - poisonset / targetset / validset

        """
        if self.args.local_rank is None:
            if self.args.poisonkey is None:
                self.init_seed = np.random.randint(0, 2**32 - 1)
            else:
                self.init_seed = int(self.args.poisonkey)
            set_random_seed(self.init_seed)
            print(f'Initializing Poison data (chosen images, examples, targets, labels) with random seed {self.init_seed}')
        else:
            rank = torch.distributed.get_rank()
            if self.args.poisonkey is None:
                init_seed = torch.randint(0, 2**32 - 1, [1], device=self.setup['device'])
            else:
                init_seed = torch.as_tensor(int(self.args.poisonkey), dtype=torch.int64, device=self.setup['device'])
            torch.distributed.broadcast(init_seed, src=0)
            if rank == 0:
                print(f'Initializing Poison data (chosen images, examples, targets, labels) with random seed {init_seed.item()}')
            self.init_seed = init_seed.item()
            set_random_seed(self.init_seed)
        # Parse threat model
        self.poison_setup = self._parse_threats_randomly()
        self.poisonset, self.targetset, self.validset = self._choose_poisons_randomly()

    def _parse_threats_randomly(self):
        """Parse the different threat models.

        The threat-models are [In order of expected difficulty]:

        single-class replicates the threat model of feature collision attacks,
        third-party draws all poisons from a class that is unrelated to both target and intended label.
        random-subset draws poison images from all classes.
        random-subset draw poison images from all classes and draws targets from different classes to which it assigns
        different labels.
        """
        num_classes = len(self.trainset.classes)

        target_class = np.random.randint(num_classes)
        list_intentions = list(range(num_classes))
        list_intentions.remove(target_class)
        intended_class = [np.random.choice(list_intentions)] * self.args.targets

        if self.args.targets < 1:
            poison_setup = dict(poison_budget=0, target_num=0,
                                poison_class=np.random.randint(num_classes), target_class=None,
                                intended_class=[np.random.randint(num_classes)])
            warnings.warn('Number of targets set to 0.')
            return poison_setup

        if self.args.threatmodel == 'single-class':
            poison_class = intended_class[0]
            poison_setup = dict(poison_budget=self.args.budget, target_num=self.args.targets,
                                poison_class=poison_class, target_class=target_class, intended_class=intended_class)
        elif self.args.threatmodel == 'third-party':
            list_intentions.remove(intended_class[0])
            poison_class = np.random.choice(list_intentions)
            poison_setup = dict(poison_budget=self.args.budget, target_num=self.args.targets,
                                poison_class=poison_class, target_class=target_class, intended_class=intended_class)
        elif self.args.threatmodel == 'self-betrayal':
            poison_class = target_class
            poison_setup = dict(poison_budget=self.args.budget, target_num=self.args.targets,
                                poison_class=poison_class, target_class=target_class, intended_class=intended_class)
        elif self.args.threatmodel == 'random-subset':
            poison_class = None
            poison_setup = dict(poison_budget=self.args.budget,
                                target_num=self.args.targets, poison_class=None, target_class=target_class,
                                intended_class=intended_class)
        elif self.args.threatmodel == 'random-subset-random-targets':
            target_class = None
            intended_class = np.random.randint(num_classes, size=self.args.targets)
            poison_class = None
            poison_setup = dict(poison_budget=self.args.budget,
                                target_num=self.args.targets, poison_class=None, target_class=None,
                                intended_class=intended_class)
        else:
            raise NotImplementedError('Unknown threat model.')

        return poison_setup

    def _choose_poisons_randomly(self):
        """Subconstruct poison and targets.

        The behavior is different for poisons and targets. We still consider poisons to be part of the original training
        set and load them via trainloader (And then add the adversarial pattern Delta)
        The targets are fully removed from the validation set and returned as a separate dataset, indicating that they
        should not be considered during clean validation using the validloader

        """
        # Poisons:
        if self.poison_setup['poison_class'] is not None:
            class_ids = []
            for index in range(len(self.trainset)):  # we actually iterate this way not to iterate over the images
                target, idx = self.trainset.get_target(index)
                if target == self.poison_setup['poison_class']:
                    class_ids.append(idx)

            poison_num = int(np.ceil(self.args.budget * len(self.trainset)))
            if len(class_ids) < poison_num:
                warnings.warn(f'Training set is too small for requested poison budget. \n'
                              f'Budget will be reduced to maximal size {len(class_ids)}')
                poison_num = len(class_ids)
            self.poison_ids = torch.tensor(np.random.choice(
                class_ids, size=poison_num, replace=False), dtype=torch.long)
        else:
            total_ids = []
            for index in range(len(self.trainset)):  # we actually iterate this way not to iterate over the images
                _, idx = self.trainset.get_target(index)
                total_ids.append(idx)
            poison_num = int(np.ceil(self.args.budget * len(self.trainset)))
            if len(total_ids) < poison_num:
                warnings.warn(f'Training set is too small for requested poison budget. \n'
                              f'Budget will be reduced to maximal size {len(total_ids)}')
                poison_num = len(total_ids)
            self.poison_ids = torch.tensor(np.random.choice(
                total_ids, size=poison_num, replace=False), dtype=torch.long)

        # Targets:
        if self.poison_setup['target_class'] is not None:
            class_ids = []
            for index in range(len(self.validset)):  # we actually iterate this way not to iterate over the images
                target, idx = self.validset.get_target(index)
                if target == self.poison_setup['target_class']:
                    class_ids.append(idx)
            self.target_ids = np.random.choice(class_ids, size=self.args.targets, replace=False)
        else:
            total_ids = []
            for index in range(len(self.validset)):  # we actually iterate this way not to iterate over the images
                _, idx = self.validset.get_target(index)
                total_ids.append(idx)
            self.target_ids = np.random.choice(total_ids, size=self.args.targets, replace=False)

        targetset = Subset(self.validset, indices=self.target_ids)
        valid_indices = []
        for index in range(len(self.validset)):
            _, idx = self.validset.get_target(index)
            if idx not in self.target_ids:
                valid_indices.append(idx)
        validset = Subset(self.validset, indices=valid_indices)
        poisonset = Subset(self.trainset, indices=self.poison_ids)

        # Construct lookup table
        self.poison_lookup = dict(zip(self.poison_ids.tolist(), range(poison_num)))
        return poisonset, targetset, validset



"""LMBD dataset wrap an existing dataset and create a database if necessary."""

import os
import io
import pickle

import platform
import lmdb

import torch
import torchvision
import numpy as np

class LMDBDataset(torch.utils.data.Dataset):
    """Implement LMDB caching and access.

    Based on https://github.com/pytorch/vision/blob/master/torchvision/datasets/lsun.py
    and
    https://github.com/Lyken17/Efficient-PyTorch/blob/master/tools/folder2lmdb.py
    """

    def __init__(self, dataset, database_path='', name='', rebuild_cache=False):
        """Initialize with a given pytorch dataset."""
        if os.path.isfile(database_path):
            raise ValueError('LMDB path must lead to a folder containing the databases.')
        self.dataset = dataset
        self.path = os.path.join(os.path.expanduser(database_path), f'{type(dataset).__name__}_{name}.lmdb')

        self.img_shape = self.dataset[0][0].shape

        if rebuild_cache:
            if os.path.isfile(self.path):
                os.remove(self.path)
                os.remove(self.path + '-lock')

        # Load or create database
        if os.path.isfile(self.path):
            print(f'Reusing cached database at {self.path}.')
        else:
            os.makedirs(os.path.expanduser(database_path), exist_ok=True)
            print(f'Creating database at {self.path}. This may take some time ...')
            checksum = create_database(self.dataset, self.path, mean=self.data_mean, std=self.data_std)

        # Setup database
        self.db = lmdb.open(self.path, subdir=False, max_readers=128, readonly=True, lock=False,
                            readahead=False, meminit=False, max_spare_txns=128)
        with self.db.begin(write=False) as txn:
            try:
                self.length = pickle.loads(txn.get(b'__len__'))
                self.keys = pickle.loads(txn.get(b'__keys__'))
            except TypeError:
                raise ValueError(f'The provided LMDB dataset at {self.path} is unfinished or damaged.')


    def __getattr__(self, name):
        """Call this only if all attributes of Subset are exhausted."""
        return getattr(self.dataset, name)

    def __len__(self):
        """Draw length from target dataset."""
        return self.length

    def __getitem__(self, index):
        """Get from database. This is still unordered access.

        Cursor access would be a major hassle in conjunction with random shuffles.
        """
        img, target, idx = None, None, None
        with self.db.begin(write=False) as txn:
            byteflow = txn.get(self.keys[index])

        # buffer magic
        buffer = io.BytesIO()
        buffer.write(byteflow)
        buffer.seek(0)

        np_img = np.reshape(np.frombuffer(buffer.read(), dtype=np.uint8), self.img_shape)
        img = torch.as_tensor(np_img, dtype=torch.float) / 255
        img = torchvision.transforms.functional.normalize(img, self.data_mean, self.data_std, inplace=True)

        # load label and id
        target, idx = self.dataset.get_target(index)

        return img, target, idx

    def get_target(self, index):
        """Return only the pair (target, idx)."""
        return self.dataset.get_target(index)


def create_database(dataset, database_path, write_frequency=5000, mean=(0, 0, 0), std=(1, 1, 1)):
    """Create an LMDB database from the given pytorch dataset.

    https://github.com/Lyken17/Efficient-PyTorch/blob/master/tools/folder2lmdb.py

    Removed pyarrow dependency
    """
    if platform.system() == 'Linux':
        map_size = 1099511627776 * 2  # Linux can grow memory as needed.
    else:
        raise ValueError('Provide a reasonable default map_size for your operating system.')
    db = lmdb.open(database_path, subdir=False,
                   map_size=map_size, readonly=False,
                   meminit=False, map_async=True)
    txn = db.begin(write=True)
    for idx, (img, target, idx) in enumerate(dataset):
        # unnormalize
        img_uint8 = (_unnormalize(img, mean, std).mul(255).add_(0.5).clamp_(0, 255)).to(torch.uint8)
        # serialize
        byteflow = img_uint8.numpy().tobytes(order='C')
        txn.put(u'{}'.format(idx).encode('ascii'), byteflow)
        if idx % write_frequency == 0:
            print(f"[{idx} / {len(dataset)}]")
            txn.commit()
            txn = db.begin(write=True)

    # finalize dataset
    txn.commit()
    keys = [u'{}'.format(k).encode('ascii') for k in range(idx + 1)]
    with db.begin(write=True) as txn:
        txn.put(b'__keys__', pickle.dumps(keys))
        txn.put(b'__len__', pickle.dumps(len(keys)))

def _unnormalize(tensor, mean, std, inplace=False):
    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
    std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
    tensor.mul_(std[:, None, None]).add_(mean[:, None, None])
    return tensor
